"""
This module contains functions and objects for working with relaxation data
"""

from . import utils as ut

import numpy as np
from scipy.optimize import curve_fit, OptimizeWarning
import sys
if sys.platform[0:3] == "win" or sys.platform == "darwin":
    import matplotlib
    matplotlib.use('TkAgg')
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider
from matplotlib.ticker import ScalarFormatter

from . import gui


def select_model_rates_plot(model: 'Relaxation_profile', xvals, rate,
                            button_options):
    """
    Creates matplotlib figure in which the user selects which type of
    process(es) they would like to fit their relaxation rate data to
    """

    model.RP_function = None

    fig, ax = plt.subplots(figsize=(6.0, 6.0), num='Select relaxation model')
    ax.set_xscale('log')
    ax.set_yscale('log')

    if model.process == 'tau_vs_T':

        supt = "Temperature dependence of relaxation rates."
        supt += "\nSelect type of process."

        fig.suptitle(supt, fontsize=10)
        # Set x and y limits
        ut.set_rate_xy_lims(ax, rate, np.abs(model.yerr_plt), xvals)
        ax.set_xlabel(r'Temperature (K)')
        # set the position of the buttons.
        button_position = [0.1, 0.35, 0.3, 0.75]

    elif model.process == 'tau_vs_H':
        supt = "Field dependence of relaxation rates."
        supt += "\nSelect type of process."

        fig.suptitle(supt, fontsize=10)
        ax.set_xlabel(r'Field (Oe)')
        # set the position of the buttons.
        button_position = [0.35, 0.63, 0.25, 0.3]  # l, b, w, h
    else:
        exit("No model process selected")

    # Add uncertainties as errorbars
    ax.errorbar(
        xvals,
        rate,
        yerr=np.abs(model.yerr_plt),
        marker='o',
        ls='none',
        fillstyle='none',
        color='r'
    )

    # Create buttons for different processes
    radio = gui.make_mpl_radiobuttons(
        button_position, button_options, 0.03, figure=fig
    )

    def model_radiobutton(radio, figure, model):
        print('    Model function {} has been selected.'.format(
            radio.value_selected
        ))
        model.RP_function = radio.value_selected

        # Stop blocking of this figure and then close
        plt.close(figure)
        fig.canvas.stop_event_loop()

    radio.on_clicked(
        lambda _: model_radiobutton(radio, fig, model)
    )

    ax.set_ylabel(r'$\tau^{-1}$ (s$^{-1}$)')

    # if user closes plot, stop blocking (see below)
    cid = fig.canvas.mpl_connect(
        'close_event',
        lambda _: fig.canvas.stop_event_loop()
    )
    # Enable interactive mode, and from now on manually control whether
    # figures block (remain open)
    plt.ion()
    plt.show()
    # Begin blocking of this figure
    fig.canvas.start_event_loop()
    # Disconnect canvas id
    fig.canvas.mpl_disconnect(cid)

    if model.RP_function is None:
        exit("No model selected")

    return


def select_model_rates(model: 'Relaxation_profile', xvals, rates, field,
                       _process=None):

    model.field = field

    if _process is not None:
        model.process = _process

    if model.process == 'tau_vs_T':

        if model.field:
            options = [
                'Orbach + Direct',
                'Raman + Direct',
                'Orbach + Raman + Direct'
            ]
        else:
            options = [
                'Orbach',
                'Raman',
                'QTM',
                'Orbach + Raman',
                'Orbach + QTM',
                'Raman + QTM',
                'Orbach + Raman + QTM'
            ]

    elif model.process == 'tau_vs_H':

        # If there is a data point measured at 0 field, raise an error.
        if float(0) in xvals:
            print('\n***Error***:')
            print('Zero-field data point present in the data.')
            print('Delete point and try again.')

        options = [
            'QTM(H) + Raman-II + Ct',
            'QTM(H) + Raman-II + Ct - constrained',
            'Brons-Van-Vleck & Raman-II',
            'Brons-Van-Vleck & Ct',
            'Brons-Van-Vleck & Ct - constrained'
        ]

    # Create plot for user interaction
    select_model_rates_plot(
        model, xvals, rates, options
    )

    # Define relaxation model and bounds attributes
    model.fitting_params(model.RP_function)

    # Fit relaxation rates with help from user
    guess_fit_rates(
        model, xvals, rates
    )

    return


def guess_fit_rates(model: 'Relaxation_profile', xvals, rate):

    fig, ax = plt.subplots(figsize=(7, 6), num='Choose starting guess')
    fig.subplots_adjust(left=0.15, bottom=0.15)
    ax.errorbar(
        xvals,
        rate,
        yerr=model.yerr_plt,
        marker='o',
        ls='none',
        fillstyle='none',
        color='r'
    )
    ax.set_xscale('log')
    ax.set_yscale('log')

    axcolor = 'lightgoldenrodyellow'

    # Initial guess for all possible params.
    # The actual value is defined as log because it is passed
    # to the 10** functions.
    _init = {
        # ([tau0, min, max], [Ueff, min, max])
        'Orbach': ([-5., 1E-13, 1E-1], [500, 5, 2000]),
        # ([C, min, max], [n, min, max])
        'Raman': ([-4, 1E-9, 1E-1], [9, 1, 12]),
        # ([tau0QTM, min, max])
        'QTM': ([-2, 1E-6, 1000], []),
        # ([A, min, max])
        'Direct': ([-5., 1E-10, 1], []),
        # ([e, f], [min, max], [min, max])
        'BronsVanVleck': ([-5., -5.], [1E-10, 1], [1E-10, 1]),
        # ([CII, mII],, [minCII, maxCII] [minmII, maxmII])
        'RamanII': ([-4, 4], [1E-30, 1E-1], [0.5, 8]),
        # ([ZFQTM, FQTM, p], [min, max], [min, max], [min, max])
        'FieldQTM': ([-2, -2, 2], [1E-30, 1000], [1E-30, 1000], [0.5, 6.]),
        # ([Ct], [min, max])
        'Cte':  ([-4], [1E-14, 1E4])
    }

    # Positions of buttons in the plot. [left, bottom, width, height]
    _buttons = {
        'Reset':  [0.91, 0.83, 0.08, 0.05],
        'Fit': [0.91, 0.75, 0.06, 0.05],
        'Change': [0.91, 0.62, 0.08, 0.08],
        # Orbach
        'tau0': [0.24, 0.890, 0.20, 0.02],
        # Orbach
        'U': [0.64, 0.890, 0.20, 0.02],
        # Raman
        'C': [0.24, 0.925, 0.20, 0.02],
        # Raman
        'n': [0.64, 0.925, 0.20, 0.02],
        # QTM
        'qtm': [0.44, 0.965, 0.20, 0.02],
        # Direct
        'A': [0.44, 0.965, 0.20, 0.02],
        # Brons-Van-Vleck
        'e': [0.22, 0.890, 0.20, 0.02],
        # Brons-Van-Vleck
        'f': [0.22, 0.925, 0.20, 0.02],
        # RamanII
        'CII': [0.68, 0.890, 0.20, 0.02],
        # RamanII
        'mII': [0.68, 0.925, 0.20, 0.02],
        # Raman (isothermal)
        'Cte': [0.68, 0.965, 0.20, 0.02],
        # QTM(H)
        'ZFQTM':  [0.22, 0.890, 0.20, 0.02],
        # QTM(H)
        'FQTM': [0.22, 0.925, 0.20, 0.02],
        # QTM(H)
        'p': [0.22, 0.965, 0.20, 0.02]
    }
    # options = [
    #     'QTM(H) + Raman-II + Cte',
    #     'Brons-Van-Vleck & Raman-II',
    #     'Brons-Van-Vleck & Cte'
    # ]

    # Define the buttons.
    resetax = plt.axes(_buttons['Reset'], figure=fig)
    fitax = plt.axes(_buttons['Fit'], figure=fig)
    changemodelax = plt.axes(_buttons['Change'], figure=fig)

    button = gui.make_mpl_button(resetax, 'Reset', 'lightgoldenrodyellow')
    button_fit = gui.make_mpl_button(fitax, 'Fit', 'lightblue')
    button_change = gui.make_mpl_button(
        changemodelax, 'Change\nmodel', 'lightcoral'
    )

    # Attempt first fit of reduced set of points to get reasonable
    # parameter estimates.
    if model.RP_function == 'Orbach':
        try:
            popt, _ = curve_fit(
                model.Orbach,
                xvals[-3:],
                np.log10(rate[-3:]),
                bounds=model.bounds
            )
            init_pre, init_U = popt
            model._guess = popt
            # Tau0 slider bounds
            minpre = 10**popt[0]*0.1
            maxpre = 10**popt[0]*10
            # Ueff slider bounds
            minU = popt[1]-popt[1]/2.
            maxU = popt[1]+popt[1]/2

        except (RuntimeError, OptimizeWarning):
            print('Could not attempt initial guess')
            print('Use sliders to sample the parameter space.')
            init_pre = _init['Orbach'][0][0]
            init_U = _init['Orbach'][1][0]
            minpre = _init['Orbach'][0][1]
            maxpre = _init['Orbach'][0][2]
            minU = _init['Orbach'][1][1]
            maxU = _init['Orbach'][1][2]

        # Plot
        plot, = ax.loglog(
            xvals,
            np.power(
                10., model.Orbach(xvals, init_pre, init_U)
            ),
            '-',
            label=model.RP_function
        )

        # Axes and sliders for parameters
        axpre = plt.axes(_buttons['tau0'], facecolor=axcolor)
        axU_1 = plt.axes(_buttons['U'], facecolor=axcolor)

        # Tau0
        spre_1 = Slider(
            axpre,
            r'$\tau_0$ (s)',
            np.log10(minpre),
            np.log10(maxpre),
            np.log10(10**init_pre),
            valfmt='%4.2E'
        )
        # Ueff
        sU_1 = Slider(
            axU_1,
            r'$U_{\mathrm{eff}}$ (K)',
            minU,
            maxU,
            init_U,
            valfmt='%3.2f'
        )

        spre_1.valtext.set_text('{:4.2E}'.format(10**spre_1.val))

        def update(slider_1, slider_2, model, plot):
            prefac, U_eff = [slider_1.val, slider_2.val]
            model._guess = [slider_1.val, slider_2.val]
            slider_1.valtext.set_text('{:4.2E}'.format(10**prefac))
            plot.set_ydata(
                np.power(10., model.Orbach(xvals, prefac, U_eff))
            )

        spre_1.on_changed(lambda _: update(spre_1, sU_1, model, plot))
        sU_1.on_changed(lambda _: update(spre_1, sU_1, model, plot))
        button.on_clicked(lambda _: gui.sliders_reset([spre_1, sU_1]))

    elif model.RP_function == 'Raman':

        try:
            popt, _ = curve_fit(
                model.Raman,
                xvals[-3:],
                np.log10(rate[-3:]),
                bounds=model.bounds
            )
            init_C, init_n = popt
            model._guess = [init_C, init_n]
            # C limits
            minC = 10**init_C*0.1
            maxC = 10**init_C*10

            # n limits
            minn = init_n-init_n/2.
            maxn = init_n+init_n/2.

        except (RuntimeError, OptimizeWarning):
            print('Could not attempt initial guess')
            print('Use sliders to sample the parameter space.')
            init_C = _init['Raman'][0][0]
            init_n = _init['Raman'][1][0]
            minC = _init['Raman'][0][1]
            maxC = _init['Raman'][0][2]
            minn = _init['Raman'][1][1]
            maxn = _init['Raman'][1][2]

        plot, = ax.loglog(
            xvals,
            np.power(10., model.Raman(xvals, init_C, init_n)),
            '-',
            label=model.RP_function
        )

        # C slider
        axC = plt.axes(_buttons['C'], facecolor=axcolor)
        sC = Slider(
            axC,
            r'$C$ (s$^{-1}$ K$^{-n}$)',
            np.log10(minC),
            np.log10(maxC),
            np.log10(10**init_C),
            valfmt='%4.2E'
        )
        sC.valtext.set_text('{:4.2E}'.format(10**sC.val))

        # n slider
        axn = plt.axes(_buttons['n'], facecolor=axcolor)
        sn = Slider(axn, '$n$', minn, maxn, init_n, valfmt='%3.2f')

        def update(slider_1, slider_2, model, plot):
            C, n = [slider_1.val, slider_2.val]
            model._guess = [slider_1.val, slider_2.val]
            slider_1.valtext.set_text('{:4.2E}'.format(10**C))
            plot.set_ydata(np.power(10., model.Raman(xvals, C, n)))

        def reset(slider_1, slider_2):
            slider_1.reset()
            slider_2.reset()

        # Update and reset plot based on slider and button input
        sC.on_changed(lambda _: update(sC, sn, model, plot))
        sn.on_changed(lambda _: update(sC, sn, model, plot))
        button.on_clicked(lambda _: reset(sC, sn))

    elif model.RP_function == 'QTM':
        try:
            popt, _ = curve_fit(
                model.QTM,
                xvals,
                np.log10(rate),
                bounds=model.bounds
            )
            init_QTM = popt[0]
            model._guess = init_QTM
            minQTM, maxQTM = 10**init_QTM*0.1, 10**init_QTM*10
        except (RuntimeError, OptimizeWarning):
            print('Could not attempt initial guess')
            print('Use sliders to sample the parameter space.')
            init_QTM = _init['QTM'][0][0]
            minQTM = _init['QTM'][0][1]
            maxQTM = _init['QTM'][0][2]

        plot, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.QTM(xvals, init_QTM)
            ),
            '-',
            label=model.RP_function
        )
        axqtm = plt.axes(_buttons['qtm'], facecolor=axcolor)
        sqtm = Slider(
            axqtm,
            r'$\tau_{\mathrm{QTM}}$ (s)',
            np.log10(minQTM),
            np.log10(maxQTM),
            np.log10(10**init_QTM),
            valfmt='%4.2E'
        )
        sqtm.valtext.set_text('{:4.2E}'.format(10**sqtm.val))

        def update(model, slider, plot):
            model._guess = [slider.val]
            slider.valtext.set_text('{:4.2E}'.format(10**slider.val))
            plot.set_ydata(np.power(10., model.QTM(xvals, slider.val)))

        sqtm.on_changed(lambda _: update(model, sqtm, plot))
        button.on_clicked(gui.sliders_reset([sqtm]))

    elif model.RP_function == 'Orbach + Raman':
        try:
            popt, _ = curve_fit(
                model.Orbach,
                xvals[-3:],
                np.log10(rate[-3:]),
                bounds=model.bounds[:, 0:2]
            )
            init_pre, init_U = popt
            popt, _ = curve_fit(
                model.Raman,
                xvals[:3],
                np.log10(rate[:3]),
                bounds=model.bounds[:, 2::]
            )
            init_C, init_n = popt
            model._guess = [init_pre, init_U, init_C, init_n]
            minpre = 10**init_pre*0.1
            maxpre = 10**init_pre*10
            minU = init_U-init_U/2.
            maxU = init_U+init_U/2.
            minC = 10**init_C*0.1
            maxC = 10**init_C*10
            minn = init_n-init_n/2.
            maxn = init_n+init_n/2.

        except (RuntimeError, OptimizeWarning):
            print('Could not attempt initial guess')
            print('Use sliders to sample the parameter space.')
            init_pre = _init['Orbach'][0][0]
            init_U = _init['Orbach'][1][0]
            minpre = _init['Orbach'][0][1]
            maxpre = _init['Orbach'][0][2]
            minU = _init['Orbach'][1][1]
            maxU = _init['Orbach'][1][2]

            init_C = _init['Raman'][0][0]
            init_n = _init['Raman'][1][0]
            minC = _init['Raman'][0][1]
            maxC = _init['Raman'][0][2]
            minn = _init['Raman'][1][1]
            maxn = _init['Raman'][1][2]

        plot_1, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.Orbach_Raman(
                    xvals,
                    init_pre,
                    init_U,
                    init_C,
                    init_n
                )
            ),
            lw=1.5,
            label=model.RP_function,
            ls='-'
        )
        plot_2, = ax.loglog(
            xvals,
            np.power(
                10., model.Orbach(xvals, init_pre, init_U)
            ),
            lw=1.3,
            label='Orbach',
            ls=':'
        )
        plot_3, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.Raman(xvals, init_C, init_n)
            ),
            lw=1.3,
            label='Raman',
            ls='--'
        )

        axpre = plt.axes(_buttons['tau0'], facecolor=axcolor)
        spre_1 = Slider(
            axpre,
            r'$\tau_0$ (s)',
            np.log10(minpre),
            np.log10(maxpre),
            np.log10(10**init_pre),
            valfmt='%4.2E'
        )
        spre_1.valtext.set_text('{:4.2E}'.format(10**spre_1.val))

        axU_1 = plt.axes(_buttons['U'], facecolor=axcolor)
        sU_1 = Slider(
            axU_1,
            r'$U_{\mathrm{eff}}$ (K)',
            minU,
            maxU,
            init_U,
            valfmt='%3.2f'
        )

        axC = plt.axes(_buttons['C'], facecolor=axcolor)
        sC = Slider(
            axC,
            r'$C$ (s$^{-1}$ K$^{-n}$)',
            np.log10(minC),
            np.log10(maxC),
            np.log10(10**init_C),
            valfmt='%4.2E'
        )
        sC.valtext.set_text('{:4.2E}'.format(10**sC.val))

        axn = plt.axes(_buttons['n'], facecolor=axcolor)
        sn = Slider(axn, r'$n$', minn, maxn, init_n, valfmt='%3.2f')

        def update(model, slider_1, slider_2, slider_3, slider_4,
                   plot_1, plot_2, plot_3):
            prefac, U_eff, C, n = [
                slider.val for slider in [
                    slider_1, slider_2, slider_3, slider_4
                ]
            ]
            model._guess = [prefac, U_eff, C, n]
            slider_1.valtext.set_text('{:4.2E}'.format(10**prefac))
            slider_3.valtext.set_text('{:4.2E}'.format(10**C))
            plot_1.set_ydata(
                np.power(10., model.Orbach_Raman(xvals, *model._guess))
            )
            plot_2.set_ydata(np.power(10., model.Orbach(xvals, prefac, U_eff)))
            plot_3.set_ydata(np.power(10., model.Raman(xvals, C, n)))

            return

        sliders = [spre_1, sU_1, sC, sn]
        plots = [plot_1, plot_2, plot_3]

        for slider in sliders:
            slider.on_changed(lambda _: update(model, *sliders, *plots))

        button.on_clicked(lambda _: gui.sliders_reset(sliders))

    elif model.RP_function == 'Orbach + QTM':
        try:
            popt, _ = curve_fit(
                model.Orbach,
                xvals[-3:],
                np.log10(rate[-3:]),
                bounds=model.bounds[:, 0:2]
            )
            init_pre, init_U = popt
            popt, _ = curve_fit(
                model.QTM,
                xvals[:3],
                np.log10(rate[:3]),
                bounds=model.bounds[:, 2::]
            )
            init_QTM = popt[0]
            model._guess = [init_pre, init_U, init_QTM]
            minpre = 10**init_pre*0.1
            maxpre = 10**init_pre*10
            minU = init_U-init_U/2.
            maxU = init_U+init_U/2.
            minQTM = 10**init_QTM*0.1
            maxQTM = 10**init_QTM*10

        except (RuntimeError, OptimizeWarning):
            print('Could not attempt initial guess')
            print('Use sliders to sample the parameter space.')
            init_pre = _init['Orbach'][0][0]
            init_U = _init['Orbach'][1][0]
            minpre = _init['Orbach'][0][1]
            maxpre = _init['Orbach'][0][2]
            minU = _init['Orbach'][1][1]
            maxU = _init['Orbach'][1][2]

            init_QTM = _init['QTM'][0][0]
            minQTM = _init['QTM'][0][1]
            maxQTM = _init['QTM'][0][2]

        plot_1, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.Orbach_QTM(xvals, init_pre, init_U, init_QTM)
            ),
            lw=1.5,
            label=model.RP_function,
            ls='-'
        )
        plot_2, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.Orbach(xvals, init_pre, init_U)
            ),
            lw=1.3,
            label='Orbach',
            ls=':'
        )
        plot_3, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.QTM(xvals, init_QTM)
            ),
            lw=1.3,
            label='QTM',
            ls='-.'
        )

        axpre = plt.axes(_buttons['tau0'], facecolor=axcolor)
        spre_1 = Slider(
            axpre,
            r'$\tau_0$ (s)',
            np.log10(minpre),
            np.log10(maxpre),
            np.log10(10**init_pre),
            valfmt='%4.2E'
        )
        spre_1.valtext.set_text('{:4.2E}'.format(10**spre_1.val))

        axU_1 = plt.axes(_buttons['U'], facecolor=axcolor)
        sU_1 = Slider(
            axU_1,
            r'$U_{\mathrm{eff}}$ (K)',
            minU,
            maxU,
            init_U,
            valfmt='%3.2f'
        )

        axqtm = plt.axes(_buttons['qtm'], facecolor=axcolor)
        sqtm = Slider(
            axqtm,
            r'$\tau_{\mathrm{QTM}}$ (s)',
            np.log10(minQTM),
            np.log10(maxQTM),
            np.log10(10**init_QTM),
            valfmt='%4.2E'
        )
        sqtm.valtext.set_text('{:4.2E}'.format(10**sqtm.val))

        def update(model, slider_1, slider_2, slider_3, plot_1, plot_2,
                   plot_3):
            prefac, U_eff, qtm = [
                slider.val for slider in [
                    slider_1, slider_2, slider_3
                ]
            ]
            model._guess = [prefac, U_eff, qtm]
            slider_1.valtext.set_text('{:4.2E}'.format(10**prefac))
            slider_3.valtext.set_text('{:4.2E}'.format(10**qtm))
            plot_1.set_ydata(
                np.power(10., model.Orbach_QTM(xvals, *model._guess))
            )
            plot_2.set_ydata(np.power(10., model.Orbach(xvals, prefac, U_eff)))
            plot_3.set_ydata(np.power(10., model.QTM(xvals, qtm)))

            return

        sliders = [spre_1, sU_1, sqtm]
        plots = [plot_1, plot_2, plot_3]

        for slider in sliders:
            slider.on_changed(lambda _: update(model, *sliders, *plots))

        button.on_clicked(lambda _: gui.sliders_reset(sliders))

    elif model.RP_function == 'Raman + QTM':
        try:
            popt, _ = curve_fit(
                model.Raman,
                xvals[-3:],
                np.log10(rate[-3:]),
                bounds=model.bounds[:, 0:2]
            )
            init_C, init_n = popt
            popt, _ = curve_fit(
                model.QTM,
                xvals[:3],
                np.log10(rate[:3]),
                bounds=model.bounds[:, -1:]
            )
            init_QTM = popt[0]
            model._guess = [init_C, init_n, init_QTM]
            minC = 10**init_C*0.1
            maxC = 10**init_C*10
            minn = init_n-init_n/2.
            maxn = init_n+init_n/2.
            minQTM = 10**init_QTM*0.1
            maxQTM = 10**init_QTM*10

        except (RuntimeError, OptimizeWarning):
            print('Could not attempt initial guess')
            print('Use sliders to sample the parameter space.')
            init_C = _init['Raman'][0][0]
            init_n = _init['Raman'][1][0]
            minC = _init['Raman'][0][1]
            maxC = _init['Raman'][0][2]
            minn = _init['Raman'][1][1]
            maxn = _init['Raman'][1][2]

            init_QTM = _init['QTM'][0][0]
            minQTM = _init['QTM'][0][1]
            maxQTM = _init['QTM'][0][2]

        plot_1, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.Raman_QTM(
                    xvals,
                    init_C,
                    init_n,
                    init_QTM
                )
            ),
            lw=1.5,
            label=model.RP_function,
            ls='-'
        )

        plot_2, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.Raman(xvals, init_C, init_n)
            ),
            lw=1.3,
            label='Raman',
            ls='--'
        )
        plot_3, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.QTM(xvals, init_QTM)
            ),
            lw=1.3,
            label='QTM',
            ls='-.'
        )

        axC = plt.axes(_buttons['C'], facecolor=axcolor)
        sC = Slider(
            axC,
            r'$C$ (s$^{-1}$ K$^{-n}$)',
            np.log10(minC),
            np.log10(maxC),
            np.log10(10**init_C),
            valfmt='%4.2E'
        )
        sC.valtext.set_text('{:4.2E}'.format(10**sC.val))

        axn = plt.axes(_buttons['n'], facecolor=axcolor)
        sn = Slider(axn, r'$n$', minn, maxn, init_n, valfmt='%3.2f')

        axqtm = plt.axes(_buttons['qtm'], facecolor=axcolor)
        sqtm = Slider(
            axqtm,
            r'$\tau_{\mathrm{QTM}}$ (s)',
            np.log10(minQTM),
            np.log10(maxQTM),
            np.log10(10**init_QTM),
            valfmt='%4.2E'
        )
        sqtm.valtext.set_text('{:4.2E}'.format(10**sqtm.val))

        def update(model, slider_1, slider_2, slider_3,
                   plot_1, plot_2, plot_3):
            C, n, qtm = [
                slider.val for slider in [
                    slider_1, slider_2, slider_3
                ]
            ]
            model._guess = [C, n, qtm]
            slider_1.valtext.set_text('{:4.2E}'.format(10**C))
            slider_3.valtext.set_text('{:4.2E}'.format(10**qtm))
            plot_1.set_ydata(
                np.power(10., model.Raman_QTM(xvals, *model._guess))
            )
            plot_2.set_ydata(np.power(10., model.Raman(xvals, C, n)))
            plot_3.set_ydata(np.power(10., model.QTM(xvals, qtm)))

            return

        sliders = [sC, sn, sqtm]
        plots = [plot_1, plot_2, plot_3]

        for slider in sliders:
            slider.on_changed(lambda _: update(model, *sliders, *plots))

        button.on_clicked(lambda _: gui.sliders_reset(sliders))

    elif model.RP_function == 'Orbach + Raman + QTM':
        try:
            popt, _ = curve_fit(
                model.Orbach,
                xvals[-3:],
                np.log10(rate[-3:]),
                bounds=model.bounds[:, 0:2]
            )
            init_pre, init_U = popt
            popt, _ = curve_fit(
                model.Raman,
                xvals[5:9],
                np.log10(rate[5:9]),
                bounds=model.bounds[:, 2:4]
            )
            init_C, init_n = popt
            popt, _ = curve_fit(
                model.QTM,
                xvals[:3],
                np.log10(rate[:3]),
                bounds=model.bounds[:, 4::]
            )
            init_QTM = popt[0]
            model._guess = [init_pre, init_U, init_C, init_n, init_QTM]
            minpre = 10**init_pre*0.1
            maxpre = 10**init_pre*10
            minU = init_U-init_U/2.
            maxU = init_U+init_U/2.
            minC = 10**init_C*0.1
            maxC = 10**init_C*10
            minn = init_n-init_n/2.
            maxn = init_n+init_n/2.
            minQTM = 10**init_QTM*0.1
            maxQTM = 10**init_QTM*10

        except (RuntimeError, OptimizeWarning):
            print('Could not attempt initial guess')
            print('Use sliders to sample the parameter space.')
            init_pre = _init['Orbach'][0][0]
            init_U = _init['Orbach'][1][0]
            minpre = _init['Orbach'][0][1]
            maxpre = _init['Orbach'][0][2]
            minU = _init['Orbach'][1][1]
            maxU = _init['Orbach'][1][2]

            init_C = _init['Raman'][0][0]
            init_n = _init['Raman'][1][0]
            minC = _init['Raman'][0][1]
            maxC = _init['Raman'][0][2]
            minn = _init['Raman'][1][1]
            maxn = _init['Raman'][1][2]

            init_QTM = _init['QTM'][0][0]
            minQTM = _init['QTM'][0][1]
            maxQTM = _init['QTM'][0][2]

        plot_1, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.Orbach_Raman_QTM(
                    xvals,
                    init_pre,
                    init_U,
                    init_C,
                    init_n,
                    init_QTM
                )
            ),
            lw=1.5,
            label=model.RP_function,
            ls='-'
        )
        plot_2, = ax.loglog(
            xvals,
            np.power(
                10., model.Orbach(xvals, init_pre, init_U)
            ),
            lw=1.3,
            label='Orbach',
            ls=':'
        )
        plot_3, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.Raman(xvals, init_C, init_n)
            ),
            lw=1.3,
            label='Raman',
            ls='--'
        )
        plot_4, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.QTM(xvals, init_QTM)
            ),
            lw=1.3,
            label='QTM',
            ls='-.'
        )
        axpre = plt.axes(_buttons['tau0'], facecolor=axcolor)
        spre_1 = Slider(
            axpre,
            r'$\tau_0$ (s)',
            np.log10(minpre),
            np.log10(maxpre),
            np.log10(10**init_pre),
            valfmt='%4.2E'
        )
        spre_1.valtext.set_text('{:4.2E}'.format(10**spre_1.val))

        axU_1 = plt.axes(_buttons['U'], facecolor=axcolor)
        sU_1 = Slider(
            axU_1,
            r'$U_{\mathrm{eff}}$ (K)',
            minU,
            maxU,
            init_U,
            valfmt='%3.2f'
        )

        axC = plt.axes(_buttons['C'], facecolor=axcolor)
        sC = Slider(
            axC,
            r'$C$ (s$^{-1}$ K$^{-n}$)',
            np.log10(minC),
            np.log10(maxC),
            np.log10(10**init_C),
            valfmt='%4.2E'
        )
        sC.valtext.set_text('{:4.2E}'.format(10**sC.val))

        axn = plt.axes(_buttons['n'], facecolor=axcolor)
        sn = Slider(axn, r'$n$', minn, maxn, init_n, valfmt='%3.2f')

        axqtm = plt.axes(_buttons['qtm'], facecolor=axcolor)
        sqtm = Slider(
            axqtm,
            r'$\tau_{\mathrm{QTM}}$ (s)',
            np.log10(minQTM),
            np.log10(maxQTM),
            np.log10(10**init_QTM),
            valfmt='%4.2E'
        )
        sqtm.valtext.set_text('{:4.2E}'.format(10**sqtm.val))

        def update(model, slider_1, slider_2, slider_3, slider_4, slider_5,
                   plot_1, plot_2, plot_3, plot_4):
            prefac, U_eff, C, n, qtm = [
                slider.val for slider in [
                    slider_1, slider_2, slider_3, slider_4, slider_5
                ]
            ]
            model._guess = [prefac, U_eff, C, n, qtm]
            slider_1.valtext.set_text('{:4.2E}'.format(10**prefac))
            slider_3.valtext.set_text('{:4.2E}'.format(10**C))
            slider_5.valtext.set_text('{:4.2E}'.format(10**qtm))
            plot_1.set_ydata(
                np.power(10., model.Orbach_Raman_QTM(xvals, *model._guess))
            )
            plot_2.set_ydata(np.power(10., model.Orbach(xvals, prefac, U_eff)))
            plot_3.set_ydata(np.power(10., model.Raman(xvals, C, n)))
            plot_4.set_ydata(np.power(10., model.QTM(xvals, qtm)))

            return

        sliders = [spre_1, sU_1, sC, sn, sqtm]
        plots = [plot_1, plot_2, plot_3, plot_4]

        for slider in sliders:
            slider.on_changed(lambda _: update(model, *sliders, *plots))

        button.on_clicked(lambda _: gui.sliders_reset(sliders))

    elif model.RP_function == 'Orbach + Direct':
        try:
            popt, _ = curve_fit(
                model.Orbach,
                xvals,
                np.log10(rate),
                bounds=[model.bounds[0, :2], model.bounds[1, :2]]
            )
            init_pre, init_U = popt
            popt, _ = curve_fit(
                model.Direct,
                xvals,
                np.log10(rate),
                bounds=[model.bounds[0, -1], model.bounds[1, -1]]
            )
            init_A = popt[0]
            model._guess = [init_pre, init_U, init_A]
            minpre = 10**init_pre*0.1
            maxpre = 10**init_pre*10
            minU = init_U-init_U/2.
            maxU = init_U+init_U/2.
            minA = 10**init_A*0.1
            maxA = 10**init_A*10
        except (RuntimeError, OptimizeWarning):
            print('Could not attempt initial guess')
            print('Use sliders to sample the parameter space.')
            init_pre = _init['Orbach'][0][0]
            init_U = _init['Orbach'][1][0]
            init_A = _init['Direct'][0][0]
            minpre = _init['Orbach'][0][1]
            maxpre = _init['Orbach'][0][2]
            minU = _init['Orbach'][1][1]
            maxU = _init['Orbach'][1][2]
            minA = _init['Direct'][0][1]
            maxA = _init['Direct'][0][2]

        # Orbach + Direct
        plot_1, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.Orbach_Direct(xvals, init_pre, init_U, init_A)
            ),
            '-',
            label=model.RP_function
        )
        # Orbach
        plot_2, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.Orbach(xvals, init_pre, init_U)
            ),
            '--',
            label='Orbach'
        )
        # Direct
        plot_3, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.Direct(xvals, init_A)
            ),
            ':',
            label='Direct'
        )
        axpre = plt.axes(_buttons['tau0'], facecolor=axcolor)
        spre_1 = Slider(
            axpre,
            r'$\tau_0$ (s)',
            np.log10(minpre),
            np.log10(maxpre),
            np.log10(10**init_pre),
            valfmt='%4.2E'
        )

        axU_1 = plt.axes(_buttons['U'], facecolor=axcolor)
        sU_1 = Slider(
            axU_1,
            r'$U_{\mathrm{eff}}$ (K)',
            minU,
            maxU,
            init_U,
            valfmt='%3.2f'
        )

        axA = plt.axes(_buttons['A'], facecolor=axcolor)
        sA = Slider(
            axA,
            r'A (K$^{-1}$s$^{-1}$)',
            np.log10(minA),
            np.log10(maxA),
            np.log10(10**init_A),
            valfmt='%4.2E'
        )
        # Slider is in log10 scale to appear linear, need to convert back
        # for text
        # Stored value is passed as exponent to model function
        spre_1.valtext.set_text('{:4.2E}'.format(10**spre_1.val))
        sA.valtext.set_text('{:4.2E}'.format(10**sA.val))

        def update(model, slider_1, slider_2, slider_3, plot_1, plot_2,
                   plot_3):
            prefac, U_eff, A = [slider_1.val, slider_2.val, slider_3.val]
            model._guess = [slider_1.val, slider_2.val, slider_3.val]
            slider_1.valtext.set_text('{:4.2E}'.format(10**prefac))
            slider_3.valtext.set_text('{:4.2E}'.format(10**A))
            plot_1.set_ydata(
                np.power(10., model.Orbach_Direct(xvals, prefac, U_eff, A))
            )
            plot_2.set_ydata(
                np.power(10., model.Orbach(xvals, prefac, U_eff))
            )
            plot_3.set_ydata(
                np.power(10., model.Direct(xvals, A))
            )

        sliders = [spre_1, sU_1, sA]
        plots = [plot_1, plot_2, plot_3]

        for slider in sliders:
            slider.on_changed(lambda _: update(model, *sliders, *plots))

        button.on_clicked(lambda _: gui.sliders_reset(sliders))

    elif model.RP_function == 'Raman + Direct':
        try:
            popt, _ = curve_fit(
                model.Raman,
                xvals,
                np.log10(rate),
                bounds=[model.bounds[0, :2], model.bounds[1, :2]]
            )
            init_C, init_n = popt
            popt, _ = curve_fit(
                model.Direct,
                xvals,
                np.log10(rate),
                bounds=[model.bounds[0, -1], model.bounds[1, -1]]
            )
            init_A = popt[0]
            model._guess = [init_C, init_n, init_A]
            minC = 10**init_C*0.1
            maxC = 10**init_C*10
            minn = init_n-init_n/2.
            maxn = init_n+init_n/2.
            minA = 10**init_A*0.1
            maxA = 10**init_A*10
        except (RuntimeError, OptimizeWarning):
            print('Could not attempt initial guess')
            print('Use sliders to sample the parameter space.')
            init_C = _init['Raman'][0][0]
            init_n = _init['Raman'][1][0]
            init_A = _init['Direct'][0][0]
            minC = _init['Raman'][0][1]
            maxC = _init['Raman'][0][2]
            minn = _init['Raman'][1][1]
            maxn = _init['Raman'][1][2]
            minA = _init['Direct'][0][1]
            maxA = _init['Direct'][0][2]

        plot_1, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.Raman_Direct(xvals, init_C, init_n, init_A)
            ),
            '-',
            label=model.RP_function
        )
        plot_2, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.Raman(xvals, init_C, init_n)
            ),
            '--',
            label='Raman'
        )
        plot_3, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.Direct(xvals, init_A)
            ),
            ':',
            label='Direct'
        )
        axC = plt.axes(_buttons['C'], facecolor=axcolor)
        sC = Slider(
            axC,
            r'$C$ (s$^{-1}$ K$^{-n}$)',
            np.log10(minC),
            np.log10(maxC),
            np.log10(10**init_C),
            valfmt='%4.2E'
        )
        sC.valtext.set_text('{:4.2E}'.format(10**sC.val))

        axn = plt.axes(_buttons['n'], facecolor=axcolor)
        sn = Slider(
            axn,
            r'$n$',
            minn,
            maxn,
            init_n,
            valfmt='%3.2f'
        )

        axA = plt.axes(_buttons['A'], facecolor=axcolor)
        sA = Slider(
            axA,
            r'A (K$^{-1}$s$^{-1}$)',
            np.log10(minA),
            np.log10(maxA),
            np.log10(10**init_A),
            valfmt='%4.2E'
        )
        sA.valtext.set_text('{:4.2E}'.format(10**sA.val))

        def update(model, slider_1, slider_2, slider_3, plot_1,
                   plot_2, plot_3):
            C, n, A = [slider_1.val, slider_2.val, slider_3.val]
            model._guess = [slider_1.val, slider_2.val, slider_3.val]
            slider_1.valtext.set_text('{:4.2E}'.format(10**C))
            slider_3.valtext.set_text('{:4.2E}'.format(10**A))
            plot_1.set_ydata(
                np.power(10., model.Raman_Direct(xvals, C, n, A))
            )
            plot_2.set_ydata(
                np.power(10., model.Raman(xvals, C, n))
            )
            plot_3.set_ydata(
                np.power(10., model.Direct(xvals, A))
            )

        sliders = [sC, sn, sA]
        plots = [plot_1, plot_2, plot_3]

        for slider in sliders:
            slider.on_changed(lambda _: update(model, *sliders, *plots))

        button.on_clicked(lambda _: gui.sliders_reset(sliders))

    elif model.RP_function == 'Orbach + Raman + Direct':
        try:
            popt, _ = curve_fit(
                model.Orbach,
                xvals,
                np.log10(rate),
                bounds=[model.bounds[0, :2], model.bounds[1, :2]]
            )
            init_pre, init_U = popt
            popt, _ = curve_fit(
                model.Raman,
                xvals,
                np.log10(rate),
                bounds=[model.bounds[0, 2:4], model.bounds[1, 2:4]]
            )
            init_C, init_n = popt
            popt, _ = curve_fit(
                model.Direct,
                xvals,
                np.log10(rate),
                bounds=[model.bounds[0, -1], model.bounds[1, -1]]
            )
            init_A = popt[0]
            model._guess = [init_pre, init_U, init_C, init_n, init_A]
            minpre = 10**init_pre*0.1
            maxpre = 10**init_pre*10
            minU = init_U-init_U/2.
            maxU = init_U+init_U/2.
            minC = 10**init_C*0.1
            maxC = 10**init_C*10
            minn = init_n-init_n/2.
            maxn = init_n+init_n/2.
            minA = 10**init_A*0.1
            maxA = 10**init_A*10

        except (RuntimeError, OptimizeWarning):
            print('Could not attempt initial guess')
            print('Use sliders to sample the parameter space.')
            init_pre = _init['Orbach'][0][0]
            init_U = _init['Orbach'][1][0]
            init_C = _init['Raman'][0][0]
            init_n = _init['Raman'][1][0]
            init_A = _init['Direct'][0][0]
            minpre = _init['Orbach'][0][1]
            maxpre = _init['Orbach'][0][2]
            minU = _init['Orbach'][1][1]
            maxU = _init['Orbach'][1][2]
            minC = _init['Raman'][0][1]
            maxC = _init['Raman'][0][2]
            minn = _init['Raman'][1][1]
            maxn = _init['Raman'][1][2]
            minA = _init['Direct'][0][1]
            maxA = _init['Direct'][0][2]

        plot_1, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.Orbach_Raman_Direct(
                    xvals, init_pre, init_U, init_C, init_n, init_A
                )
            ),
            '-',
            label=model.RP_function
        )
        plot_2, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.Orbach(xvals, init_pre, init_U)
            ),
            '--',
            label='Orbach'
        )
        plot_3, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.Raman(xvals, init_C, init_n)
            ),
            ':',
            label='Raman'
        )
        plot_4, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.Direct(xvals, init_A)
            ),
            '-.',
            label='Direct'
        )
        axpre = plt.axes(_buttons['tau0'], facecolor=axcolor)
        spre_1 = Slider(
            axpre,
            r'$\tau_0$ (s)',
            np.log10(minpre),
            np.log10(maxpre),
            np.log10(10**init_pre),
            valfmt='%4.2E'
        )
        spre_1.valtext.set_text('{:4.2E}'.format(10**spre_1.val))

        axC = plt.axes(_buttons['C'], facecolor=axcolor)
        sC = Slider(
            axC,
            r'$C$ (s$^{-1}$ K$^{-n}$)',
            np.log10(minC),
            np.log10(maxC),
            np.log10(10**init_C),
            valfmt='%4.2E'
        )
        sC.valtext.set_text('{:4.2E}'.format(10**sC.val))

        axn = plt.axes(_buttons['n'], facecolor=axcolor)
        sn = Slider(
            axn,
            r'$n$',
            minn,
            maxn,
            init_n,
            valfmt='%3.2f'
        )

        axA = plt.axes(_buttons['A'], facecolor=axcolor)
        sA = Slider(
            axA,
            r'A (K$^{-1}$s$^{-1}$)',
            np.log10(minA),
            np.log10(maxA),
            np.log10(10**init_A),
            valfmt='%4.2E'
        )
        sA.valtext.set_text('{:4.2E}'.format(10**sA.val))

        axU_1 = plt.axes(_buttons['U'], facecolor=axcolor)
        sU_1 = Slider(
            axU_1,
            r'$U_{\mathrm{eff}}$ (K)',
            minU,
            maxU,
            init_U,
            valfmt='%3.2f'
        )

        def update(model, slider_1, slider_2, slider_3, slider_4, slider_5,
                   plot_1, plot_2, plot_3, plot_4):
            prefac, U_eff, C, n, A = [
                slider_1.val, slider_2.val, slider_3.val, slider_4.val,
                slider_5.val
            ]
            model._guess = [
                slider_1.val, slider_2.val, slider_3.val, slider_4.val,
                slider_5.val
            ]
            slider_1.valtext.set_text('{:4.2E}'.format(10**prefac))
            slider_3.valtext.set_text('{:4.2E}'.format(10**slider_3.val))
            slider_5.valtext.set_text('{:4.2E}'.format(10**A))
            plot_1.set_ydata(
                np.power(
                    10.,
                    model.Orbach_Raman_Direct(
                        xvals, prefac, U_eff, C, n, A
                    )
                )
            )
            plot_2.set_ydata(np.power(10., model.Orbach(xvals, prefac, U_eff)))
            plot_3.set_ydata(np.power(10., model.Raman(xvals, C, n)))
            plot_4.set_ydata(np.power(10., model.Direct(xvals, A)))

        sliders = [
            spre_1, sU_1, sC, sn, sA
        ]

        plots = [plot_1, plot_2, plot_3, plot_4]

        for slider in sliders:
            slider.on_changed(lambda _: update(model, *sliders, *plots))

        button.on_clicked(lambda _: gui.sliders_reset(sliders))

    elif model.RP_function in ['Brons-Van-Vleck & Ct', 'Brons-Van-Vleck & Ct - constrained']: # noqa

        plot, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.BronsVanVleck_Cte(
                    xvals,
                    *(_init['BronsVanVleck'][0] + _init['Cte'][0])
                )
            ),
            lw=1.5,
            label=model.RP_function,
            ls='-'
        )

        axe = plt.axes(_buttons['e'], facecolor=axcolor)

        se = Slider(
            axe,
            r'$e$ (Oe$^{-2}$)',
            np.log10(_init['BronsVanVleck'][1][0]),
            np.log10(_init['BronsVanVleck'][1][1]),
            np.log10(10**_init['BronsVanVleck'][0][0]),
            valfmt='%4.2E'
            )
        se.valtext.set_text('{:4.2E}'.format(10**se.val))

        axf = plt.axes(_buttons['f'], facecolor=axcolor)
        sf = Slider(
            axf,
            r'$f$ (Oe$^{-2}$)',
            np.log10(_init['BronsVanVleck'][2][0]),
            np.log10(_init['BronsVanVleck'][2][1]),
            np.log10(10**_init['BronsVanVleck'][0][1]),
            valfmt='%4.2E'
        )
        sf.valtext.set_text('{:4.2E}'.format(10**sf.val))

        axcte = plt.axes(_buttons['Cte'], facecolor=axcolor)
        scte = Slider(
            axcte,
            r'$Ct$',
            np.log10(_init['Cte'][1][0]),
            np.log10(_init['Cte'][1][1]),
            np.log10(10**_init['Cte'][0][0]),
            valfmt='%4.2E'
        )
        scte.valtext.set_text('{:4.2E}'.format(10**scte.val))

        def update(model, slider_1, slider_2, slider_3, plot):
            if model.RP_function == 'Brons-Van-Vleck & Ct':
                model._guess = [slider_1.val, slider_2.val, slider_3.val]
            elif model.RP_function == 'Brons-Van-Vleck & Ct - constrained':
                model._guess = [slider_1.val, slider_2.val]
                model.Ct_fixed = slider_3.val
            slider_1.valtext.set_text('{:4.2E}'.format(10**slider_1.val))
            slider_2.valtext.set_text('{:4.2E}'.format(10**slider_2.val))
            slider_3.valtext.set_text('{:4.2E}'.format(10**slider_3.val))
            plot.set_ydata(
                np.power(10., model.BronsVanVleck_Cte(
                    xvals, slider_1.val, slider_2.val, slider_3.val
                ))
            )

        sliders = [se, sf, scte]

        for slider in sliders:
            slider.on_changed(lambda _: update(model, *sliders, plot))

        button.on_clicked(gui.sliders_reset(sliders))

    elif model.RP_function == 'Brons-Van-Vleck & Raman-II':

        plot, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.BronsVanVleck_RamanII(
                    xvals,
                    *(_init['BronsVanVleck'][0] + _init['RamanII'][0])
                )
            ),
            lw=1.5,
            label=model.RP_function,
            ls='-'
        )

        axe = plt.axes(_buttons['e'], facecolor=axcolor)
        se = Slider(
            axe,
            r'$e$ (Oe$^{-2}$)',
            np.log10(_init['BronsVanVleck'][1][0]),
            np.log10(_init['BronsVanVleck'][1][1]),
            np.log10(10**_init['BronsVanVleck'][0][0]),
            valfmt='%4.2E'
        )
        se.valtext.set_text('{:4.2E}'.format(10**se.val))

        axf = plt.axes(_buttons['f'], facecolor=axcolor)
        sf = Slider(
            axf,
            r'$f$ (Oe$^{-2}$)',
            np.log10(_init['BronsVanVleck'][2][0]),
            np.log10(_init['BronsVanVleck'][2][1]),
            np.log10(10**_init['BronsVanVleck'][0][1]),
            valfmt='%4.2E'
        )
        sf.valtext.set_text('{:4.2E}'.format(10**sf.val))

        axCII = plt.axes(_buttons['CII'], facecolor=axcolor)
        sCII = Slider(
            axCII,
            r'$C$ (s$^{-1}$ Oe$^{-m}$)',
            np.log10(_init['RamanII'][1][0]),
            np.log10(_init['RamanII'][1][1]),
            np.log10(10**_init['RamanII'][0][0]),
            valfmt='%4.2E'
        )
        sCII.valtext.set_text('{:4.2E}'.format(10**sCII.val))

        axmII = plt.axes(_buttons['mII'], facecolor=axcolor)
        smII = Slider(
            axmII,
            r'$m$',
            _init['RamanII'][2][0],
            _init['RamanII'][2][1],
            _init['RamanII'][0][1],
            valfmt='%3.2f'
        )

        def update(model, slider_1, slider_2, slider_3, slider_4, plot):
            model._guess = [
                slider_1.val, slider_2.val, slider_3.val, slider_4.val
            ]
            slider_1.valtext.set_text('{:4.2E}'.format(10**slider_1.val))
            slider_2.valtext.set_text('{:4.2E}'.format(10**slider_2.val))
            slider_3.valtext.set_text('{:4.2E}'.format(10**slider_3.val))
            plot.set_ydata(
                np.power(
                    10.,
                    model.BronsVanVleck_RamanII(xvals, *model._guess)
                )
            )

        sliders = [se, sf, sCII, smII]

        for slider in sliders:
            slider.on_changed(lambda _: update(model, *sliders, plot))

        button.on_clicked(gui.sliders_reset(sliders))

    elif model.RP_function in ['QTM(H) + Raman-II + Ct', 'QTM(H) + Raman-II + Ct - constrained']: # noqa

        plot_1, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.FieldQTM_RamanII_Ct(
                    xvals,
                    *(_init['FieldQTM'][0] + _init['RamanII'][0] + _init['Cte'][0]) # noqa
                )
            ),
            lw=1.5,
            label=model.RP_function,
            ls='-'
        )
        plot_2, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.FieldQTM(
                    xvals, *_init['FieldQTM'][0])
            ),
            lw=1.3,
            label='QTM(H)',
            ls=':'
        )
        plot_3, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.RamanII(
                    xvals, *_init['RamanII'][0]
                )
            ),
            lw=1.3,
            label='Raman-II',
            ls='--'
        )
        plot_4, = ax.loglog(
            xvals,
            np.power(
                10.,
                model.Ct(xvals, *_init['Cte'][0])
            ),
            lw=1.3,
            label='Ct',
            ls='-.'
        )

        axzfqtm = plt.axes(_buttons['ZFQTM'], facecolor=axcolor)
        szfqtm = Slider(
            axzfqtm,
            r'$\tau_{\mathrm{QTM}}$ (s)',
            np.log10(_init['FieldQTM'][1][0]),
            np.log10(_init['FieldQTM'][1][1]),
            np.log10(10**_init['FieldQTM'][0][0]),
            valfmt='%4.2E'
        )
        szfqtm.valtext.set_text('{:4.2E}'.format(10**szfqtm.val))

        axfqtm = plt.axes(_buttons['FQTM'], facecolor=axcolor)
        sfqtm = Slider(
            axfqtm,
            r'$\tau_{\mathrm{QTM(H)}}$ (Oe$^{-p}$)',
            np.log10(_init['FieldQTM'][2][0]),
            np.log10(_init['FieldQTM'][2][1]),
            np.log10(10**_init['FieldQTM'][0][1]),
            valfmt='%4.2E'
        )
        sfqtm.valtext.set_text('{:4.2E}'.format(10**sfqtm.val))

        axp = plt.axes(_buttons['p'], facecolor=axcolor)
        sp = Slider(
            axp,
            r'$p$',
            _init['FieldQTM'][3][0],
            _init['FieldQTM'][3][1],
            _init['FieldQTM'][0][2],
            valfmt='%3.2f'
        )

        axCII = plt.axes(_buttons['CII'], facecolor=axcolor)
        sCII = Slider(
            axCII,
            r'$C$ (s$^{-1}$ Oe$^{-m}$)',
            np.log10(_init['RamanII'][1][0]),
            np.log10(_init['RamanII'][1][1]),
            np.log10(10**_init['RamanII'][0][0]),
            valfmt='%4.2E'
        )
        sCII.valtext.set_text('{:4.2E}'.format(10**sCII.val))

        axmII = plt.axes(_buttons['mII'], facecolor=axcolor)
        smII = Slider(
            axmII,
            r'$m$', _init['RamanII'][2][0],
            _init['RamanII'][2][1],
            _init['RamanII'][0][1],
            valfmt='%3.2f'
        )

        axcte = plt.axes(_buttons['Cte'], facecolor=axcolor)
        scte = Slider(
            axcte,
            r'$Ct$',
            np.log10(_init['Cte'][1][0]),
            np.log10(_init['Cte'][1][1]),
            np.log10(10**_init['Cte'][0][0]),
            valfmt='%4.2E'
        )
        scte.valtext.set_text('{:4.2E}'.format(10**scte.val))

        def update(model, sliders, plot_1, plot_2, plot_3, plot_4):
            if model.RP_function == 'QTM(H) + Raman-II + Ct':
                model._guess = [slider.val for slider in sliders]
            elif model.RP_function == 'QTM(H) + Raman-II + Ct - constrained':
                model._guess = [slider.val for slider in sliders[1:-1]]
                model.ZFQTM_fixed = sliders[0].val
                model.Ct_fixed = sliders[-1].val

            sliders[0].valtext.set_text('{:4.2E}'.format(10**sliders[0].val))
            sliders[1].valtext.set_text('{:4.2E}'.format(10**sliders[1].val))
            sliders[3].valtext.set_text('{:4.2E}'.format(10**sliders[3].val))
            sliders[5].valtext.set_text('{:4.2E}'.format(10**sliders[5].val))
            plot_1.set_ydata(
                np.power(
                    10.,
                    model.FieldQTM_RamanII_Ct(
                        xvals,
                        *[slider.val for slider in sliders]
                    )
                )
            )
            plot_2.set_ydata(
                np.power(
                    10.,
                    model.FieldQTM(xvals, szfqtm.val, sfqtm.val, sp.val)
                )
            )
            plot_3.set_ydata(
                np.power(10., model.RamanII(xvals, sCII.val, smII.val))
            )
            plot_4.set_ydata(np.power(10., model.Ct(xvals, scte.val)))

            return

        plots = [plot_1, plot_2, plot_3, plot_4]
        sliders = [szfqtm, sfqtm, sp, sCII, smII, scte]

        for slider in sliders:
            slider.on_changed(lambda _: update(model, sliders, *plots))

        button.on_clicked(gui.sliders_reset(sliders))

    if model.process == 'tau_vs_T':
        # Set x and y limits
        ut.set_rate_xy_lims(ax, rate, model.yerr_plt, xvals)
        # Reset x limits and formatter
        ut.set_rate_x_lims(ax, xvals)
        # Set xlabel
        ax.set_xlabel(r'Temperature (K)')
    elif model.process == 'tau_vs_H':
        # Set xlabel
        ax.set_xlabel(r'Field (Oe)')

    ax.set_ylabel(r'$\tau^{-1}$ (s$^{-1}$)')
    ax.legend(loc=0, fontsize='10', numpoints=1, ncol=2, frameon=False)

    def close_then_fit(model, xvals, rate, figure, cid):

        figure.canvas.mpl_disconnect(cid)
        plt.close()
        figure.canvas.stop_event_loop()

        fit_rates(model, xvals, rate)

        return

    def choose_model_again(model, xvals, rate, figure, cid):

        figure.canvas.mpl_disconnect(cid)
        plt.close()
        figure.canvas.stop_event_loop()
        plt.ioff()

        print("Resetting")

        select_model_rates(
            model, xvals, rate, field=model.field, _process=model.process
        )

        return

    # When clicking the Fit button.
    button_fit.on_clicked(
        lambda _: close_then_fit(model, xvals, rate, fig, cid)
    )
    # When clicking the Change Model button.
    button_change.on_clicked(
        lambda _: choose_model_again(model, xvals, rate, fig, cid)
    )
    # If user exits plot, stop blocking
    cid = fig.canvas.mpl_connect(
        'close_event',
        lambda _: fig.canvas.stop_event_loop()
    )

    fig.canvas.start_event_loop()
    plt.show(block=False)
    fig.canvas.mpl_disconnect(cid)

    return


def fit_rates(model: 'Relaxation_profile', xvals: list[float],
              rates: list[float]):

    # Stop manual handling of event loop and blocking
    # plots from this point no longer have buttons

    try:
        # Fit the data and calculate error of fit for each parameter
        popt, pcov = curve_fit(
            model.RP_model,
            xvals,
            np.log10(rates),
            p0=model._guess,
            sigma=model.yerr_fit,
            absolute_sigma=True,
            bounds=model.bounds
        )
        perr = (np.sqrt(np.diag(pcov)))

    # If fitting is not successful, take user back to initial guess
    except (ValueError, RuntimeError):
        print(
            '\n  Specified model function {} cannot fit the data.'.format(
                model.RP_function
            )
        )
        print('    Try again.')
        select_model_rates(model, xvals, rates, field=model.field)
        return

    # Define the output files.
    # Must be done here as model.RP_function gets updated by the user
    # and this is the last instance.
    if model.exe_mode != 'relaxation':
        RP_filename, RP_params, RP_fit = ut.create_output_files(
            model.results_dir, _field=model.field, model_RP=model.RP_function,
            _output='all', exe_mode=model.exe_mode
        )
    else:
        RP_filename, RP_params, RP_fit = ut.create_output_files(
            model.results_dir, model_RP=model.RP_function,
            exe_mode=model.exe_mode,
        )

    # Create new figure and axis
    fig2, ax = plt.subplots(figsize=(6, 6), num='Relaxation rate fit')

    # Plot error bar from cole-cole fit
    ax.errorbar(
        xvals,
        rates,
        yerr=model.yerr_plt,
        marker='o',
        ls='none',
        fillstyle='none',
        color='r'
    )

    # Set log scale
    ax.set_xscale('log')
    ax.set_yscale('log')

    if model.process == 'tau_vs_T':
        # Set x and y limits
        ut.set_rate_xy_lims(ax, rates, model.yerr_plt, xvals)
        # Reset x limits and formatter
        ut.set_rate_x_lims(ax, xvals)
        # Set xlabel
        ax.set_xlabel(r'Temperature (K)')
    elif model.process == 'tau_vs_H':
        # Set xlabel
        ax.set_xlabel(r'Field (Oe)')

    # xvals for fitted eqn to be plotted at
    xvals_fit = np.linspace(xvals[0], xvals[-1], len(xvals)*10)

    ax.plot(
        xvals_fit,
        np.power(
            10,
            model.RP_model(
                xvals_fit, *popt
            )
        ),
        lw=1.5,
        label='Fit'
    )

    print(
        '    For {} process, the fitted parameters and 1-sigma'.format(
            model.RP_function
        ),
        ' estimates are are:'
    )

    if model.RP_function == 'Orbach':
        print('    log_10[t_0] = {} +- {} log_10[s]'.format(
            popt[0], perr[0]
        ))
        print('    U_eff = {} +- {} K'.format(popt[1], perr[1]))

    elif model.RP_function == 'Raman':
        print('    log_10[C] = {} +- {} log_10[s^-1 K^-n]'.format(
            popt[0], perr[0]
        ))
        print('    n = {} +- {}'.format(popt[1], perr[1]))

    elif model.RP_function == 'QTM':
        print('    log_10[t_QTM] = {} +- {} log_10[s]'.format(
            popt[0], perr[0]
        ))

    elif model.RP_function == 'Orbach + Raman':

        print('    log_10[t_0] = {} +- {} log_10[s]'.format(
            popt[0], perr[0]
        ))
        print('    U_eff = {} +- {} K'.format(popt[1], perr[1]))
        print('    log_10[C] = {} +- {} log_10[s^-1 K^-n]'.format(
            popt[2], perr[2]
        ))
        print('    n = {} +- {}'.format(popt[3], perr[3]))

        ax.plot(
            xvals_fit,
            np.power(10, model.Orbach(xvals_fit, popt[0], popt[1])),
            lw=1.3,
            label='Orbach',
            ls=':'
        )
        ax.plot(
            xvals_fit,
            np.power(10, model.Raman(xvals_fit, popt[2], popt[3])),
            lw=1.3,
            label='Raman',
            ls='--'
        )

    elif model.RP_function == 'Orbach + QTM':

        print('    log_10[t_0] = {} +- {} log_10[s]'.format(
            popt[0], perr[0]
        ))
        print('    U_eff = {} +- {} K'.format(popt[1], perr[1]))
        print('    log_10[t_QTM] = {} +- {} log_10[s]'.format(
            popt[2], perr[2]
        ))

        ax.plot(
            xvals_fit,
            np.power(10, model.Orbach(xvals_fit, popt[0], popt[1])),
            lw=1.3,
            label='Orbach',
            ls=':'
        )
        ax.plot(
            xvals_fit,
            np.power(10, model.QTM(xvals_fit, popt[2])),
            lw=1.3,
            label='QTM',
            ls='-.'
        )

    elif model.RP_function == 'Raman + QTM':

        print('    log_10[C] = {} +- {} log_10[s^-1 K^-n]'.format(popt[0], perr[0])) # noqa
        print('    n = {} +- {}'.format(popt[1], perr[1]))
        print('    log_10[t_QTM] = {} +- {} log_10[s]'.format(popt[2], perr[2])) # noqa

        ax.plot(xvals_fit, np.power(10, model.Raman(xvals_fit, popt[0],popt[1])), lw=1.3, label='Raman', ls='--') # noqa
        ax.plot(xvals_fit, np.power(10, model.QTM(xvals_fit, popt[2])), lw=1.3, label='QTM', ls='-.') # noqa

    elif model.RP_function == 'Orbach + Raman + QTM':

        print('    log_10[t_0] = {} +- {} log_10[s]'.format(popt[0], perr[0]))
        print('    U_eff = {} +- {} K'.format(popt[1], perr[1]))
        print('    log_10[C] = {} +- {} log_10[s^-1 K^-n]'.format(popt[2], perr[2])) # noqa
        print('    n = {} +- {}'.format(popt[3], perr[3]))
        print('    log_10[t_QTM] = {} +- {} log_10[s]'.format(popt[4], perr[4])) # noqa

        ax.plot(xvals_fit, np.power(10, model.Orbach(xvals_fit, popt[0],popt[1])), lw=1.3, label='Orbach', ls=':') # noqa
        ax.plot(xvals_fit, np.power(10, model.Raman(xvals_fit, popt[2],popt[3])), lw=1.3, label='Raman', ls='--') # noqa
        ax.plot(xvals_fit, np.power(10, model.QTM(xvals_fit, popt[4])), lw=1.3, label='QTM', ls='-.') # noqa

    elif model.RP_function == 'Orbach + Direct':

        print('    log_10[t_0] = {} +- {} log_10[s]'.format(popt[0], perr[0]))
        print('    U_eff = {} +- {} K'.format(popt[1], perr[1]))
        print('    log_10[A] = {} +- {} log_10[s^-1 K^-1]'.format(popt[2], perr[2])) # noqa

        ax.plot(xvals_fit, np.power(10, model.Orbach(xvals_fit, popt[0],popt[1])), lw=1.3, label='Orbach', ls=':') # noqa
        ax.plot(xvals_fit, np.power(10, model.Direct(xvals_fit, popt[2])), lw=1.3, label='Direct', ls='--') # noqa

    elif model.RP_function == 'Raman + Direct':

        print('    log_10[C] = {} +- {} log_10[s^-1 K^-n]'.format(popt[0], perr[0])) # noqa
        print('    n = {} +- {}'.format(popt[1], perr[1]))
        print('    log_10[A] = {} +- {} log_10[s^-1 K^-1]'.format(popt[2], perr[2])) # noqa

        ax.plot(xvals_fit, np.power(10, model.Raman(xvals_fit, popt[0],popt[1])), lw=1.3, label='Raman', ls=':') # noqa
        ax.plot(xvals_fit, np.power(10, model.Direct(xvals_fit, popt[2])), lw=1.3, label='Direct', ls='--') # noqa

    elif model.RP_function == 'Orbach + Raman + Direct':

        print('    log_10[t_0] = {} +- {} log_10[s]'.format(popt[0], perr[0]))
        print('    U_eff = {} +- {} K'.format(popt[1], perr[1]))
        print('    log_10[C] = {} +- {} log_10[s^-1 K^-n]'.format(popt[2], perr[2])) # noqa
        print('    n = {} +- {}'.format(popt[3], perr[3]))
        print('    log_10[A] = {} +- {} log_10[s^-1 K^-1]'.format(popt[4], perr[4])) # noqa

        ax.plot(xvals_fit, np.power(10, model.Orbach(xvals_fit, popt[0],popt[1])), lw=1.3, label='Orbach', ls=':') # noqa
        ax.plot(xvals_fit, np.power(10, model.Raman(xvals_fit, popt[2],popt[3])), lw=1.3, label='Raman', ls='--') # noqa
        ax.plot(xvals_fit, np.power(10, model.Direct(xvals_fit, popt[4])), lw=1.3, label='Direct', ls='-.') # noqa

    elif model.RP_function == 'QTM(H) + Raman-II + Ct':

        print('    log_10[t_QTM] = {} +- {} log_10[s]'.format(popt[0], perr[0])) # noqa
        print('    log_10[t_QTM(H)] = {} +- {} log_10[Oe^-p]'.format(popt[1], perr[1])) # noqa
        print('    p = {} +- {} '.format(popt[2], perr[2]))
        print('    log_10[CII] = {} +- {} log_10[s^-1 Oe^-m]'.format(popt[3], perr[3])) # noqa
        print('    m = {} +- {}'.format(popt[4], perr[4]))
        print('    log_10[Ct] = {} +- {} log_10[s^-1]'.format(popt[5], perr[5])) # noqa

        ax.plot(xvals_fit, np.power(10, model.FieldQTM(xvals_fit, popt[0], popt[1], popt[2])), lw=1.3, label='QTM(H)', ls=':') # noqa
        ax.plot(xvals_fit, np.power(10, model.RamanII(xvals_fit, popt[3], popt[4])), lw=1.3, label='RamanII', ls='--') # noqa
        ax.plot(xvals_fit, np.power(10, model.Ct(xvals_fit, popt[5])), lw=1.3, label='Ct', ls='-.') # noqa

    elif model.RP_function == 'QTM(H) + Raman-II + Ct - constrained':

        print('    log_10[t_QTM] = {} log_10[s]'.format(model.ZFQTM_fixed))
        print('    log_10[t_QTM(H)] = {} +- {} log_10[Oe^-p]'.format(popt[0], perr[0])) # noqa
        print('    p = {} +- {} '.format(popt[1], perr[1]))
        print('    log_10[CII] = {} +- {} log_10[s^-1 Oe^-m]'.format(popt[2], perr[2])) # noqa
        print('    m = {} +- {}'.format(popt[3], perr[3]))
        print('    log_10[Ct] = {} log_10[s^-1]'.format(model.Ct_fixed))

        ax.plot(xvals_fit, np.power(10, model.FieldQTM(xvals_fit, model.ZFQTM_fixed, popt[0], popt[1])), lw=1.3, label='QTM(H)', ls=':') # noqa
        ax.plot(xvals_fit, np.power(10, model.RamanII(xvals_fit, popt[2], popt[3])), lw=1.3, label='RamanII', ls='--') # noqa
        ax.plot(xvals_fit, np.power(10, model.Ct(xvals_fit, model.Ct_fixed)), lw=1.3, label='Ct', ls='-.') # noqa

    elif model.RP_function == 'Brons-Van-Vleck & Raman-II':

        print('    log_10[e] = {} +- {} log_10[(Oe^-2)]'.format(popt[0], perr[0])) # noqa
        print('    log_10[f] = {} +- {} log_10[(Oe^-2)]'.format(popt[1], perr[1])) # noqa
        print('    log_10[CII] = {} +- {} log_10[s^-1 Oe^-m]'.format(popt[2], perr[2])) # noqa
        print('    m = {} +- {}'.format(popt[3], perr[3]))

    elif model.RP_function == 'Brons-Van-Vleck & Ct':

        print('    log_10[e] = {} +- {} log_10[(Oe^-2)]'.format(popt[0], perr[0])) # noqa
        print('    log_10[f] = {} +- {} log_10[(Oe^-2)]'.format(popt[1], perr[1])) # noqa
        print('    log_10[Ct] = {} +- {} log_10[s^-1]'.format(popt[2], perr[2])) # noqa

    elif model.RP_function == 'Brons-Van-Vleck & Ct - constrained':

        print('    log_10[e] = {} +- {} log_10[(Oe^-2)]'.format(popt[0], perr[0])) # noqa
        print('    log_10[f] = {} +- {} log_10[(Oe^-2)]'.format(popt[1], perr[1])) # noqa
        print('    log_10[Ct] = {} log_10[s^-1]'.format(model.Ct_fixed))

    # Create the final expression to be written into the plot.
    expression = model.final_expression(popt, perr)
    ax.text(0.0, 1.02, s=expression, fontsize=10, transform=ax.transAxes)

    ax.legend(loc=0, fontsize='medium', numpoints=1, ncol=2, frameon=False)
    ax.set_ylabel(r'$\tau^{-1}$ (s$^{-1}$)')

    cid = fig2.canvas.mpl_connect(
        'close_event',
        lambda _: fig2.canvas.stop_event_loop()
    )
    plt.savefig('{}_fit.png'.format(RP_filename), dpi=300)
    plt.ion()
    plt.show()
    fig2.canvas.start_event_loop()
    fig2.canvas.mpl_disconnect(cid)

    print('\n    Plot of relaxation profile fit saved as')
    print('     {}_fit.png'.format(RP_filename))

    model.plot_residuals(xvals, rates, popt, model.RP_model, RP_filename)

    model._write_results(
        xvals_fit,
        10**model.RP_model(xvals_fit, *popt),
        popt,
        perr,
        RP_filename,
        RP_params,
        RP_fit
    )

    if model.verbose:
        chi_2 = np.sum(
            (np.log10(rates) - model.RP_model(xvals, *popt))/model.yerr_fit
        )**2
        print('\nThe covariance matrix is: \n', pcov)
        pcor = np.ones((len(pcov), len(pcov)))
        for i in range(len(pcov)):
            for j in range(len(pcov)):
                if i == j:
                    continue
                pcor[i, j] = pcov[i, j]/(np.sqrt(np.diag(pcov)[i]*np.diag(pcov)[j])) # noqa
        print('\nThe correlation matrix is:\n{}\n'.format(pcor))
        print('\nChi squared value is: {:.f}'.format(chi_2))

    return


class Relaxation_profile():

    def __init__(self, verbose=False, exe_mode='relaxation',
                 sigma=None, results_dir=None):
        self._guess = None
        # Name (string) of the selected option.
        self.RP_function = None
        # Actual method used in the fit.
        self.RP_model = None
        self.yerr_plt = None
        self.yerr_fit = None
        self.bounds = None
        self.field = None
        self.process = None
        # In case constrained functions for tau_vs_H are requested.
        self.ZFQTM_fixed = None
        self.Ct_fixed = None
        self.verbose = verbose
        self.exe_mode = exe_mode
        self.sigma = sigma
        self.results_dir = results_dir

    def read_rates(self, file, exe_mode, AC_model=None, DC_model=None,
                   sigma=0.):
        '''
        Reads-in the data and defines self.yerr_plt, self.yerr_fit.
        AC_model=[Debye, Generalised], DC_model=[Single, Stretched]
        '''

        if exe_mode.lower() == 'ac':

            # Read in the file
            data = np.loadtxt(file, skiprows=1)
            if len(data.shape) == 1:
                data = np.array([data])

            if AC_model == 'Debye':

                xvals = data[:, 0]
                tau = data[:, 1]
                tau_err = data[:, 2]
                tau = np.array(
                    [x for _, x in sorted(zip(xvals, tau))]
                )
                tau_err = np.array(
                    [x for _, x in sorted(zip(xvals, tau_err))]
                )
                xvals.sort()
                self.yerr_plt, self.yerr_fit = self.calculate_yerr(
                    tau, tau_err
                )

            elif 'Generalised' in AC_model:

                xvals = data[:, 0]
                tau = data[:, 1]
                tau_err = data[:, 4]
                alpha = data[:, 9]
                tau = np.array(
                    [x for _, x in sorted(zip(xvals, tau))]
                )
                tau_err = np.array(
                    [x for _, x in sorted(zip(xvals, tau_err))]
                )
                alpha = np.array(
                    [x for _, x in sorted(zip(xvals, alpha))]
                )
                xvals.sort()
                self.yerr_plt, self.yerr_fit = self.calculate_yerr(
                    tau, tau_err, _alpha=alpha, _sigma=sigma
                )

        elif exe_mode.lower() == 'dc':

            # Read in the file
            data = np.loadtxt(file, skiprows=1)

            if DC_model == 'single':

                xvals = data[:, 0]
                tau = data[:, 4]
                tau_err = data[:, 5]
                tau = np.array(
                    [x for _, x in sorted(zip(xvals, tau))]
                )
                tau_err = np.array(
                    [x for _, x in sorted(zip(xvals, tau_err))]
                    )
                xvals.sort()
                self.yerr_plt, self.yerr_fit = self.calculate_yerr(
                    tau, tau_err
                )

            elif DC_model == 'stretched':

                xvals = data[:, 0]
                tau = data[:, 4]
                tau_err = data[:, 5]
                beta = data[:, 6]
                tau = np.array(
                    [x for _, x in sorted(zip(xvals, tau))]
                )
                tau_err = np.array(
                    [x for _, x in sorted(zip(xvals, tau_err))]
                    )
                beta = np.array(
                    [x for _, x in sorted(zip(xvals, beta))]
                    )
                xvals.sort()
                self.yerr_plt, self.yerr_fit = self.calculate_yerr(
                    tau, tau_err, _beta=beta
                )

        # relaxation
        else:

            AC, DC = False, False
            xvals_AC, xvals_DC, tau_AC, tau_DC = [], [], [], []
            tau_err_AC, tau_err_DC, alpha, beta = [], [], [], []

            with open(file, 'r') as f:
                for line in f:
                    if 'AC' in line:
                        AC = True
                        # skip headers
                        # To avoid problems with spaces
                        # Only use length of the data
                        line = next(f)

                        for line in f:
                            try:
                                xvals_AC.append(float(line.split()[0]))
                                tau_AC.append(float(line.split()[1]))
                                tau_err_AC.append(float(line.split()[2]))
                                # Alpha
                                if len(line.split()) == 4:
                                    if float(line.split()[3]) == 0.:
                                        alpha.append(0.001)
                                    else:
                                        alpha.append(float(line.split()[3]))
                            except:  # TODO specify exceptions
                                break

            with open(file, 'r') as f:
                for line in f:
                    if line.find('DC') != -1:
                        DC = True
                        # skip headers
                        # To avoid problems with spaces
                        # Only use length of the data
                        line = next(f)

                        for line in f:

                            try:
                                xvals_DC.append(float(line.split()[0]))
                                tau_DC.append(float(line.split()[1]))
                                tau_err_DC.append(float(line.split()[2]))
                                # contains alpha
                                if len(line.split()) == 4:
                                    if float(line.split()[3]) == 1.:
                                        beta.append(0.999)
                                    else:
                                        beta.append(float(line.split()[3]))
                            except:  # TODO specify exceptions
                                break

            if AC and not DC:

                tau_AC = np.array(
                    [x for _, x in sorted(zip(xvals_AC, tau_AC))]
                )
                tau_err_AC = np.array(
                    [x for _, x in sorted(zip(xvals_AC, tau_err_AC))]
                )

                if len(alpha) != 0:
                    alpha = np.array(
                        [x for _, x in sorted(zip(xvals_AC, alpha))]
                    )
                else:
                    alpha = None

                xvals_AC.sort()

                self.yerr_plt, self.yerr_fit = self.calculate_yerr(
                    tau_AC, tau_err_AC, _alpha=alpha, _sigma=sigma
                )

                xvals, tau = xvals_AC, tau_AC

            elif DC and not AC:

                tau_DC = np.array(
                    [x for _, x in sorted(zip(xvals_DC, tau_DC))]
                )
                tau_err_DC = np.array(
                    [x for _, x in sorted(zip(xvals_DC, tau_err_DC))]
                )
                if len(beta) != 0:
                    beta = np.array(
                        [x for _, x in sorted(zip(xvals_DC, beta))]
                    )
                else:
                    beta = None

                xvals_DC.sort()

                self.yerr_plt, self.yerr_fit = self.calculate_yerr(
                    tau_DC, tau_err_DC, _beta=beta
                )

                xvals, tau = xvals_DC, tau_DC

            elif AC and DC:

                tau_AC = np.array(
                    [x for _, x in sorted(zip(xvals_AC, tau_AC))]
                )
                tau_err_AC = np.array(
                    [x for _, x in sorted(zip(xvals_AC, tau_err_AC))]
                )

                if len(alpha) != 0:
                    alpha = np.array(
                        [x for _, x in sorted(zip(xvals_AC, alpha))]
                    )
                else:
                    alpha = None

                xvals_AC.sort()

                tau_DC = np.array(
                    [x for _, x in sorted(zip(xvals_DC, tau_DC))]
                )
                tau_err_DC = np.array(
                    [x for _, x in sorted(zip(xvals_DC, tau_err_DC))]
                )
                if len(beta) != 0:
                    beta = np.array(
                        [x for _, x in sorted(zip(xvals_DC, beta))]
                    )
                else:
                    beta = None

                xvals_DC.sort()

                yerr_plt_AC, yerr_fit_AC = self.calculate_yerr(
                    tau_AC, tau_err_AC, _alpha=alpha, _sigma=sigma
                )
                yerr_plt_DC, yerr_fit_DC = self.calculate_yerr(
                    tau_DC, tau_err_DC, _beta=beta
                )

                # Stack them back together
                self.yerr_plt = np.hstack((yerr_plt_AC, yerr_plt_DC))
                self.yerr_fit = np.hstack((yerr_fit_AC, yerr_fit_DC))

                xvals = np.hstack((xvals_AC, xvals_DC))
                tau = np.hstack((tau_AC, tau_DC))

                # Sort all the arrays by xvals.
                # For this xvals has to be sorted last
                tau = tau[np.argsort(xvals)]

                for i in range(2):
                    self.yerr_plt[i] = self.yerr_plt[i][np.argsort(xvals)]
                self.yerr_fit = self.yerr_fit[np.argsort(xvals)]

                xvals = np.sort(xvals)

            else:
                exit("Cannot find data in specified file")

        return np.array(xvals), np.array(tau)

    def calculate_yerr(self, tau, tau_err, _alpha=None, _sigma=None,
                       _beta=None):

        if _alpha is None and _beta is None:

            yerr_plt = np.vstack(
                (
                    (1./tau - 1./(tau + tau_err)),
                    (1./(tau - tau_err) - 1./tau)
                )
            )
            # First row contains the lower errors, the second row contains the
            # upper errors.
            # this error is passed to the fitting function that returns the
            # log10 of the data
            yerr_fit = np.amax(
                (
                    abs(
                        np.log10(1./tau) - np.log10(1./(tau + tau_err))
                    ),
                    abs(
                        np.log10(1./tau) - np.log10(1./(tau - tau_err))
                    )
                ),
                axis=0
            )

            return yerr_plt, yerr_fit

        else:

            N_err_lower = 1./(tau+tau_err)
            N_err_upper = 1./(tau-tau_err)

            if _alpha is not None:

                # Recalculate tau_lognormal_err in case the user provides a
                # different sigma
                _err_lower = 1./self.log_normal(tau, _alpha, _sigma)
                _err_upper = 1./self.log_normal(tau, _alpha, -_sigma)

            if _beta is not None:

                tau_upper, tau_lower = self.Johnston(tau, _beta)

                _err_lower = 1./tau_upper
                _err_upper = 1./tau_lower

            yerr_fit = np.amax(
                (
                    abs(np.log10(1./tau) - np.log10(_err_lower)),
                    abs(np.log10(1./tau) - np.log10(_err_upper)),
                    abs(np.log10(1./tau) - np.log10(N_err_lower)),
                    abs(np.log10(1./tau) - np.log10(N_err_upper))
                ),
                axis=0
            )

            _ind = np.argmax(
                (
                    abs(np.log10(1./tau) - np.log10(_err_lower)),
                    abs(np.log10(1./tau) - np.log10(_err_upper)),
                    abs(np.log10(1./tau) - np.log10(N_err_lower)),
                    abs(np.log10(1./tau) - np.log10(N_err_upper))
                ),
                axis=0
            )

            yerr_plt = np.zeros((2, len(_ind)))
            for it, val in enumerate(_ind):
                if (val == 0 or val == 1):
                    yerr_plt[0, it] = (1./tau[it] - _err_lower[it])
                    yerr_plt[1, it] = (_err_upper[it] - 1./tau[it])
                elif (val == 2 or val == 3):
                    yerr_plt[0, it] = (1./tau[it] - N_err_lower[it])
                    yerr_plt[1, it] = (N_err_upper[it] - 1./tau[it])

            if self.verbose:
                print('Sigma used: {}'.format(_sigma))
                print('\nThe rates are:\n{}'.format(1./tau))
                print(
                    'The lower and upper errors',
                    ' in the rates are:\n{}\n'.format(yerr_plt)
                )

            return yerr_plt, yerr_fit

    def log_normal(self, tau, alpha, sigma):
        '''
        In log10 space, the positive and negative deviations from tau are
        equal -> Only the positive value is returned
        '''
        return tau*np.exp((1.82*sigma*np.sqrt(alpha))/(1-alpha))

    def Johnston(self, tau, beta):

        tau_upper = tau*np.exp(
            (1.64*np.tan(np.pi*(1-beta)/2.)) / ((1-beta)**0.141)
        )
        tau_lower = tau*np.exp(
            -(1.64*np.tan(np.pi*(1-beta)/2.))/((1-beta)**0.141)
        )

        return tau_upper, tau_lower

    # tau_vs_T under zero-field.
    @staticmethod
    def Orbach(temperature, pre_1, U_1):
        return np.log10(10**(-pre_1)*np.exp(-U_1/(temperature)))

    def Raman(self, temperature, C, n):
        return np.log10(10**C*(temperature**n))

    def QTM(self, temperature, QTM):
        return np.log10((10**(-QTM)*temperature/temperature))

    def Orbach_QTM(self, temperature, pre_1, U_1, QTM):
        # return np.log10(
        #   10**self.Orbach(temperature, pre_1, U_1)
        #   + 10**self.QTM(temperature, QTM)
        # )
        return np.log10(10**(-pre_1)*np.exp(-U_1/(temperature)) + 10**(-QTM))

    def Orbach_Raman(self, temperature, pre_1, U_1, C, n):
        # return np.log10(10**self.Orbach(temperature, pre_1, U_1) + 10**self.Raman(temperature, C, n)) # noqa
        return np.log10(
            10**(-pre_1)*np.exp(-U_1/(temperature)) + 10**C*(temperature**n)
        )

    def Orbach_Raman_QTM(self, temperature, pre_1, U_1, C, n, QTM):
        # return np.log10(10**self.Orbach(temperature, pre_1, U_1) + 10**self.Raman(temperature, C, n) + 10**self.QTM(temperature, QTM)) # noqa
        return np.log10(
            10**(-pre_1)*np.exp(-U_1/(temperature))
            + 10**C*(temperature**n)
            + 10**(-QTM)
        )

    def Raman_QTM(self, temperature, C, n, QTM):
        # return np.log10(
        #     10**self.Raman(temperature, C, n)
        #     + 10**self.QTM(temperature, QTM)
        #     )
        return np.log10(10**C*(temperature**n) + 10**(-QTM))

    def Two_Orbach(self, temperature, pre_1, U_1, pre_2, U_2):
        # return np.log10(10**self.Orbach(temperature, pre_1, U_1) + 10**self.Orbach(temperature, pre_2, U_2)) # noqa
        return np.log10(
            10**(-pre_1)*np.exp(-U_1/(temperature))
            + 10**(-pre_2)*np.exp(-U_2/(temperature))
        )

    # tau_vs_T in-field (> 0.5 T).
    def Direct(self, temperature, A):
        argu = 10**A*temperature
        return np.log10(argu)

    def Orbach_Direct(self, temperature, pre_1, U_1, A):
        argu = 10**(-pre_1)*np.exp(-U_1/(temperature))
        argu += 10**A*temperature
        return np.log10(argu)
        # return np.log10(
        # 10**self.Orbach(temperature, pre_1, U_1)
        # + 10**self.Direct(temperature, A)
        # )

    def Raman_Direct(self, temperature, C, n, A):
        argu = 10**C*(temperature**n)
        argu += 10**A*temperature
        return np.log10(argu)
        # return np.log10(
        # 10**self.Raman(temperature, C, n)
        # + 10**self.Direct(temperature, A)
        # )

    def Orbach_Raman_Direct(self, temperature, pre_1, U_1, C, n, A):
        argu = 10**(-pre_1)*np.exp(-U_1/(temperature))
        argu += 10**C*(temperature**n)
        argu += 10**A*temperature
        return np.log10(argu)
        # return np.log10
        # 10**self.Orbach(temperature, pre_1, U_1)
        # + 10**self.Raman(temperature, C, n)
        # + 10**self.Direct(temperature, A)
        # )

    # tau_vs_H.
    # ZFQTM, FQTM are the zero-field and in-field QTM params.
    def FieldQTM(self, field, ZFQTM, FQTM, p):
        return np.log10(10**(-ZFQTM)/(1 + (10**(-FQTM))*(field**p)))

    def RamanII(self, field, CII, mII):
        return np.log10((10**CII)*(field**mII))

    def Ct(self, field, Cte):
        return np.log10((10**Cte*(field/field)))

    def FieldQTM_RamanII_Ct(self, field, ZFQTM, FQTM, p, CII, mII, Cte):

        ret_val = 10**(-ZFQTM)/(1 + (10**(-FQTM))*(field**p))
        ret_val += (10**CII)*(field**mII) + 10**Cte

        return np.log10(ret_val)

    def FieldQTM_RamanII_Ct_constrained(self, field, FQTM, p, CII, mII):

        ret_val = 10**(-self.ZFQTM_fixed)/(1 + (10**(-FQTM))*(field**p))
        ret_val += (10**CII)*(field**mII) + 10**self.Ct_fixed

        return np.log10(ret_val)

    def BronsVanVleck_RamanII(self, field, e, f, CII, mII):

        ret_val = (1 + (10**e)*(field**2))/(1 + (10**f)*(field**2))
        ret_val *= 10**CII*(field**mII)

        return np.log10(ret_val)

    def BronsVanVleck_Cte(self, field, e, f, Cte):

        ret_val = (1 + (10**e)*(field**2))/(1 + (10**f)*(field**2))
        ret_val *= 10**Cte
        return np.log10(ret_val)

    def BronsVanVleck_Cte_constrained(self, field, e, f):

        ret_val = (1 + (10**e)*(field**2))/(1 + (10**f)*(field**2))
        ret_val *= 10**self.Ct_fixed
        return np.log10(ret_val)

    def fitting_params(self, RP_function):

        # Model function passed to curve_fit.

        # tau_vs_T under zero-field.
        if RP_function == 'Orbach':
            self.RP_model = self.Orbach
            # tau0, U
            self.bounds = np.array(([-30., 0.0], [np.inf, np.inf]))
        elif RP_function == 'Raman':
            self.RP_model = self.Raman
            # C, n
            self.bounds = np.array(([-30., 0.0], [np.inf, np.inf]))
        elif RP_function == 'QTM':
            self.RP_model = self.QTM
            # tau_0_QTM
            self.bounds = np.array(([-30.], [np.inf]))
        elif RP_function == 'Orbach + QTM':
            self.RP_model = self.Orbach_QTM
            # tau0, U, tau_0_QTM
            self.bounds = np.array(
                (
                    [-30., 0.0, -30.],
                    [np.inf, np.inf, np.inf]
                )
            )
        elif RP_function == 'Raman + QTM':
            self.RP_model = self.Raman_QTM
            # C, n, tau_0_QTM
            self.bounds = np.array(
                (
                    [-30., 0.0, -30.],
                    [np.inf, np.inf, np.inf]
                )
            )
        elif RP_function == 'Orbach + Raman':
            self.RP_model = self.Orbach_Raman
            # tau0, U, C, n
            self.bounds = np.array(
                (
                    [-30., 0.0, -30., 0.0],
                    [np.inf, np.inf, np.inf, np.inf]
                )
            )
        elif RP_function == 'Orbach + Raman + QTM':
            self.RP_model = self.Orbach_Raman_QTM
            # tau0, U, C, n, tau_0_QTM
            self.bounds = np.array(
                (
                    [-30., 0.0, -30., 0.0, -30.],
                    [np.inf, np.inf, np.inf, np.inf, np.inf]
                )
            )
        # tau_vs_T in-field (> 0.5 T).
        elif RP_function == 'Orbach + Direct':
            self.RP_model = self.Orbach_Direct
            # tau0, U, A
            self.bounds = np.array(
                (
                    [-30., 0.0, -30.],
                    [np.inf, np.inf, np.inf]
                )
            )
        elif RP_function == 'Raman + Direct':
            self.RP_model = self.Raman_Direct
            # C, n, A
            self.bounds = np.array(
                ([-30., 0.0, -30.], [np.inf, np.inf, np.inf]))
        elif RP_function == 'Orbach + Raman + Direct':
            self.RP_model = self.Orbach_Raman_Direct
            # tau0, U, C, n, A
            self.bounds = np.array(
                (
                    [-30., 0., -30., 0., -30.],
                    [np.inf, np.inf, np.inf, np.inf, np.inf]
                )
            )
        # tau_vs_H
        elif RP_function == 'Brons-Van-Vleck & Ct':
            self.RP_model = self.BronsVanVleck_Cte
            # e, f, Ct
            self.bounds = np.array(
                (
                    [-30., -30., -30.],
                    [np.inf, np.inf, np.inf]
                )
            )
        elif RP_function == 'Brons-Van-Vleck & Ct - constrained':
            self.RP_model = self.BronsVanVleck_Cte_constrained
            # e, f
            self.bounds = np.array(([-30., -30.], [np.inf, np.inf]))
        elif RP_function == 'Brons-Van-Vleck & Raman-II':
            self.RP_model = self.BronsVanVleck_RamanII
            # e, f, C, m
            self.bounds = np.array(
                (
                    [-30., -30., -30., 0.0],
                    [np.inf, np.inf, np.inf, 10.]
                )
            )
        elif RP_function == 'Raman-II':
            self.RP_model = self.RamanII
            # C, m
            self.bounds = np.array(([-30., 0.0], [np.inf, 10.]))
        elif RP_function == 'QTM(H)':
            self.RP_model = self.FieldQTM
            # ZFQTM, FQTM, p
            self.bounds = np.array(([-30., -30., 0.0], [np.inf, np.inf, 10.]))
        elif RP_function == 'QTM(H) + Raman-II + Ct':
            self.RP_model = self.FieldQTM_RamanII_Ct
            # ZFQTM, FQTM, p, C, m, Ct
            self.bounds = np.array(
                (
                    [-30., -30., 0.0, -30., 0.0, -30.],
                    [np.inf, np.inf, 10., np.inf, np.inf, 10.]
                )
            )
        elif RP_function == 'QTM(H) + Raman-II + Ct - constrained':
            self.RP_model = self.FieldQTM_RamanII_Ct_constrained
            # FQTM, p, C, m
            self.bounds = np.array(
                (
                    [-30., 0.0, -30., 0.0],
                    [np.inf, 10., np.inf, 10.]
                )
            )

        return

    def final_expression(self, popt, perr):

        opt_list = []
        err_list = []
        for opt, err in zip(popt, perr):
            if int('{:1.0e}'.format(err).split('e')[1]) < 0:
                err_list.append('{:1.0e}'.format(err).split('e')[0])
                opt_list.append(round(opt, abs(int('{:1.0e}'.format(err).split('e')[1])))) # noqa
            else:
                err_list.append(str(round(err, -len(str(err).split('.')[0])+1)).split('.')[0]) # noqa
                opt_list.append(str(round(opt, -len(str(err).split('.')[0])+1)).split('.')[0]) # noqa

        if self.RP_function == 'Orbach':
            expression = r'$U_{{\mathrm{{eff}}}}$ = {} ({}) K, $\tau_0$ = 10$^{{{} ({})}}$ s'.format(opt_list[1], err_list[1], opt_list[0], err_list[0]) # noqa

        elif self.RP_function == 'Raman':
            expression = r'$C$ = 10$^{{ {}({})}}$ s$^{{-1}}$ K$^{{-n}}$, $n$ = {} ({})'.format(opt_list[0], err_list[0], opt_list[1], err_list[1]) # noqa

        elif self.RP_function == 'QTM':
            expression = r'$\tau_{{\mathrm{{QTM}}}}$ = 10$^{{{} ({})}}$ s'.format(opt_list[0], err_list[0]) # noqa

        elif self.RP_function == 'Orbach + Raman':
            expression = r'$U_{{\mathrm{{eff}}}}$ = {} ({}) K, $\tau_0$ = 10$^{{{} ({})}}$ s'.format(opt_list[1], err_list[1], opt_list[0], err_list[0]) # noqa
            expression += '\n'
            expression += r'$C$ = 10$^{{ {}({})}}$ s$^{{-1}}$ K$^{{-n}}$, $n$ = {} ({})'.format(opt_list[2], err_list[2], opt_list[3], err_list[3]) # noqa

        elif self.RP_function == 'Orbach + QTM':
            expression = r'$U_{{\mathrm{{eff}}}}$ = {} ({}) K, $\tau_0$ = 10$^{{{} ({})}}$ s, $\tau_{{\mathrm{{QTM}}}}$ = 10$^{{{} ({})}}$ s'.format(opt_list[1], err_list[1], opt_list[0], err_list[0], opt_list[2], err_list[2]) # noqa

        elif self.RP_function == 'Raman + QTM':
            expression = r'$C$ = 10$^{{ {}({})}}$ s$^{{-1}}$ K$^{{-n}}$, $n$ = {} ({}), $\tau_{{\mathrm{{QTM}}}}$ = 10$^{{{} ({})}}$ s'.format(opt_list[0], err_list[0], opt_list[1], err_list[1], opt_list[2], err_list[2]) # noqa

        elif self.RP_function == 'Orbach + Raman + QTM':
            expression = r'$U_{{\mathrm{{eff}}}}$ = {} ({}) K, $\tau_0$ = 10$^{{{} ({})}}$ s'.format(opt_list[1], err_list[1], opt_list[0], err_list[0]) # noqa
            expression += '\n'
            expression += r'$C$ = 10$^{{ {}({})}}$ s$^{{-1}}$ K$^{{-n}}$, $n$ = {} ({}), $\tau_{{\mathrm{{QTM}}}}$ = 10$^{{{} ({})}}$ s'.format(opt_list[2], err_list[2], opt_list[3], err_list[3], opt_list[4], err_list[4]) # noqa

        elif self.RP_function == 'Orbach + Direct':
            expression = r'$U_{{\mathrm{{eff}}}}$ = {} ({}) K, $\tau_0$ = 10$^{{{} ({})}}$ s, $A$ = 10$^{{{} ({})}}$ s$^{{-1}}$ K$^{{-1}}$'.format( # noqa
                opt_list[1], err_list[1],
                opt_list[0], err_list[0],
                opt_list[2], err_list[2]
            )

        elif self.RP_function == 'Raman + Direct':
            expression = r'$C$ = 10$^{{ {}({})}}$ s$^{{-1}}$ K$^{{-n}}$, $n$ = {} ({}), $A$ = 10$^{{{} ({})}}$ s$^{{-1}}$ K$^{{-1}}$'.format( # noqa
                opt_list[0], err_list[0],
                opt_list[1], err_list[1],
                opt_list[2], err_list[2]
            )

        elif self.RP_function == 'Orbach + Raman + Direct':
            expression = r'$U_{{\mathrm{{eff}}}}$ = {} ({}) K, $\tau_0$ = 10$^{{{} ({})}}$ s'.format( # noqa
                opt_list[1], err_list[1],
                opt_list[0], err_list[0]
            )
            expression += '\n'
            expression += r'$C$ = 10$^{{ {}({})}}$ s$^{{-1}}$ K$^{{-n}}$, $n$ = {} ({}), $A$ = 10$^{{{} ({})}}$ s$^{{-1}}$ K$^{{-1}}$'.format( # noqa
                opt_list[2], err_list[2],
                opt_list[3], err_list[3],
                opt_list[4], err_list[4]
            )

        elif self.RP_function == 'QTM(H) + Raman-II + Ct':
            expression = r'$\tau_{{\mathrm{{QTM}}}}$ = 10$^{{{} ({})}}$ s, $\tau_{{\mathrm{{QTM(H)}}}}$ = 10$^{{{} ({})}}$ Oe$^{{-p}}$, $p$ = {} ({})'.format( # noqa
                opt_list[0], err_list[0],
                opt_list[1], err_list[1],
                opt_list[2], err_list[2]
            )
            expression += '\n'
            expression += r'$CII$ = 10$^{{ {}({})}}$ s$^{{-1}}$ Oe$^{{-m}}$, $m$ = {} ({}), $Ct$ = 10$^{{{} ({})}}$ s $^{{-1}}$'.format( # noqa
                opt_list[3], err_list[3],
                opt_list[4], err_list[4],
                opt_list[5], err_list[5]
            )

        elif self.RP_function == 'QTM(H) + Raman-II + Ct - constrained':
            expression = r'$\tau_{{\mathrm{{QTM}}}}$ = 10$^{{{:3.2f}}}$ s, $\tau_{{\mathrm{{QTM(H)}}}}$ = 10$^{{{} ({})}}$ Oe$^{{-p}}$, $p$ = {} ({})'.format( # noqa
                self.ZFQTM_fixed,
                opt_list[0], err_list[0],
                opt_list[1], err_list[1]
            )
            expression += '\n'
            expression += r'$CII$ = 10$^{{ {}({})}}$ s$^{{-1}}$ Oe$^{{-m}}$, $m$ = {} ({}), $Ct$ = 10$^{{{:3.2f}}}$ s $^{{-1}}$'.format( # noqa
                opt_list[2], err_list[2],
                opt_list[3], err_list[3],
                self.Ct_fixed
            )

        elif self.RP_function == 'Brons-Van-Vleck & Raman-II':
            expression = r'$e$ = 10$^{{{} ({})}}$ Oe$^{{-2}}$, $f$ = 10$^{{{} ({})}}$ Oe$^{{-2}}$'.format(opt_list[0], err_list[0], opt_list[1], err_list[1]) # noqa
            expression += '\n'
            expression += r'$CII$ = 10$^{{{} ({})}}$ s$^{{-1}}$ Oe$^{{-m}}$, $m$ = {} ({})'.format(opt_list[2], err_list[2], opt_list[3], err_list[3]) # noqa

        elif self.RP_function == 'Brons-Van-Vleck & Ct':
            expression = r'$e$ = 10$^{{{} ({})}}$ Oe$^{{-2}}$, $f$ = 10$^{{{} ({})}}$ Oe$^{{-2}}$'.format(opt_list[0], err_list[0], opt_list[1], err_list[1]) # noqa
            expression += '\n'
            expression += r'$Ct$ = 10$^{{{} ({})}}$ s $^{{-1}}$'.format(opt_list[2], err_list[2]) # noqa

        elif self.RP_function == 'Brons-Van-Vleck & Ct - constrained':
            expression = r'$e$ = 10$^{{{} ({})}}$ Oe$^{{-2}}$, $f$ = 10$^{{{} ({})}}$ Oe$^{{-2}}$'.format(opt_list[0], err_list[0], opt_list[1], err_list[1]) # noqa
            expression += '\n'
            expression += r'$Ct$ = 10$^{{{:3.2f}}}$ s $^{{-1}}$'.format(self.Ct_fixed) # noqa

        return expression

    def plot_residuals(self, xvals, rate, popt, model_func, RP_filename):

        # Create figure and axes
        fig, ax1 = plt.subplots(
            1,
            1,
            figsize=[6, 6],
            num='Residuals'
        )

        # Add additional set of axes to create "zero" line
        ax2 = ax1.twiny()
        ax2.get_yaxis().set_visible(False)
        ax2.set_yticks([])
        ax2.set_xticks([])
        ax2.spines["bottom"].set_position(("zero"))

        # Plot residuals
        ax1.errorbar(
            xvals,
            np.log10(rate) - model_func(xvals, *popt),
            self.yerr_fit, fmt='b.'
        )

        # Set log scale on x axis
        ax1.set_xscale('log')

        # Set formatting for y axis ticks
        ax1.yaxis.set_major_formatter(ScalarFormatter())

        # Symmetrise y axis limits based on max abs error
        all_data = [np.log10(rate) - model_func(xvals, *popt) + self.yerr_fit]
        all_data += [np.log10(rate) - model_func(xvals, *popt) - self.yerr_fit]
        ticks, maxval = ut.min_max_ticks_with_zero(all_data, 5)
        ax1.set_yticks(ticks)
        ax1.set_ylim([-maxval*1.1, +maxval*1.1])

        # Axis labels
        ax1.set_ylabel(
            r'log$_{10}[\tau^{-1}_{\mathrm{exp}}]$ - log$_{10}[\tau^{-1}_{\mathrm{fit}}]$' # noqa
        )
        if self.process == 'tau_vs_T':
            ax1.set_xlabel(r'Temperature (K)')
            # Set x limits
            ut.set_rate_x_lims(ax1, xvals)
        elif self.process == 'tau_vs_H':
            ax1.set_xlabel(r'Field (Oe)')

        # Adjust to fit y axis label
        fig.subplots_adjust(left=0.175)

        # Save plot
        cid = fig.canvas.mpl_connect(
            'close_event', lambda _: fig.canvas.stop_event_loop()
        )
        plot_file = "{}_fit_residuals.png".format(RP_filename)
        plt.savefig(plot_file, dpi=300)
        plt.ion()
        plt.show()
        fig.canvas.start_event_loop()
        fig.canvas.mpl_disconnect(cid)
        plt.ioff()

        print('\n    Plot of fit residuals saved as\n     {}'.format(
            plot_file
        ))

        return

    def _write_results(self, _temp, _fit, _popt, _perr, RP_filename, RP_params,
                       RP_fit):

        # Write the result to RP_params.
        if self.RP_function == 'QTM(H) + Raman-II + Ct - constrained':
            RP_params.write('{: 12.8E} {:^12}'.format(self.ZFQTM_fixed, '-'))
            for (i, j) in zip(_popt, _perr):
                RP_params.write('{: 12.8E} {: 12.8E}'.format(i, j))
            RP_params.write('{: 12.8E} {:^12}'.format(self.Ct_fixed, '-'))

        elif self.RP_function == 'Brons-Van-Vleck & Ct - constrained':
            RP_params.write('{: 12.8E} {:^12}'.format(self.Ct_fixed, '-'))
            for (i, j) in zip(_popt, _perr):
                RP_params.write('{: 12.8E} {: 12.8E}'.format(i, j))

        else:
            for (i, j) in zip(_popt, _perr):
                RP_params.write('{: 12.8E} {: 12.8E}'.format(i, j))

        # Write the result to RP_fit
        for (i, j) in zip(_temp, _fit):
            RP_fit.write('{: 8.6f} {: 12.8E}'.format(i, j)+'\n')

        print('\n    Parameters written to')
        print('     {}'.format(RP_filename+'_params.out'))
        print('    Fit written to')
        print('     {}'.format(RP_filename+'_fit.out'))

        return



"""This is the command line interface to cc-fit2"""
from . import ac, dc
from . import waveform as wfrm
from . import relaxation as relax
from . import utils as ut
from . import gui

import argparse
import numpy as np
import os
import requests
import sys

import warnings
warnings.filterwarnings("ignore", ".*GUI is implemented.*")
warnings.filterwarnings("ignore", "invalid value encountered in power")
warnings.filterwarnings("ignore", 'invalid value encountered in log10')
warnings.filterwarnings("ignore", 'invalid value encountered in divide')


def ac_mode_func(args):
    """
    Wrapper function for command line interface call to ac mode

    Parameters
    ----------
    args : argparser object
        command line arguments

    Returns
    -------
    None

    """

    # Object to store user configuration
    user_cfg = ut.UserConfig()

    user_cfg.file_name = args.input_file

    # mass, mw
    user_cfg.mass = args.mass
    user_cfg.mw = args.MW

    # Get file headers
    header_indices, header_names = ac.parse_headers(user_cfg.file_name)

    # check all headers are there
    if len(header_indices.values()) != 6:
        message = '\n***Error***:'
        message += '\nHeaders in the input file are not recognised.'
        message += '\nRefer to manual to see valid options, and contact'
        message += ' danielreta1@gmail.com to implement new headers.'
        if args.ccfit2_exe:
            gui.error_box(message)
        else:
            sys.exit(message)

    # Load data from file as unordered list of individual Measurements
    ac_measurements = ac.Measurement.from_file(
        user_cfg.file_name, header_indices, data_header=args.data_pointer
    )

    # Change of units of input susceptibility to cm^3mol^(-1)
    if header_names["sus_im"] in ['AC X" (emu/Oe)', "AC X'' (emu/Oe)"]:
        in_susc_units = "emu/Oe"
    else:
        in_susc_units = "emu"

    for measurement in ac_measurements:
        measurement.convert_susc(
            in_susc_units, user_cfg.mw, user_cfg.mass
        )

    # Group measurements into sublists of experiments
    # Where first dimension uses dc field strength
    # and second uses temperature
    all_experiments_sep = ac.Experiment.from_measurements(
        ac_measurements, temp_thresh=args.temp_thresh,
        field_thresh=args.field_thresh
    )

    if args.verbose:
        print('\nVerbose mode entered.\n')
        for experiments in all_experiments_sep:
            print('Field: {} Oe'.format(experiments[0].dc_field))
            print('The raw temperatures at this field have been grouped as:') # noqa
            for experiment in experiments:
                print(
                    'Mean={:04.1f}, Min={}, Max={}\nRaw={}'.format(
                        experiment.temperatures[0],
                        np.min(experiment.raw_temperatures),
                        np.max(experiment.raw_temperatures),
                        experiment.raw_temperatures
                    )
                )
                print('{} Frequencies: {}\n'.format(
                    len(experiment.ac_freq), experiment.ac_freq)
                )

    # Manual selection of temperatures to fit
    if args.select_T:
        all_experiments_sep = [
            ac.interactive_t_select(experiments)
            for experiments in all_experiments_sep
        ]
        args.discard_off = True

    if args.process == "plot":
        for experiments in all_experiments_sep:
            ac.Experiment.plot_colecole(
                experiments,
                save_dir=user_cfg.results_dir,
                file_head="ac"
            )
            ac.Experiment.plot_susceptibility(
                experiments,
                save_dir=user_cfg.results_dir,
                file_head="ac"
            )
        exit()

    unique_dc_fields = np.unique(
        [experiment[0].dc_field for experiment in all_experiments_sep]
    )

    # Eventual list of all models separated by dc_field
    all_models_sep = []

    # For each experiment, fit a model to the data
    for experiments, dc_field in zip(all_experiments_sep, unique_dc_fields):

        # Creates models for current experiments
        models = ac.Model.from_experiments(experiments, verbose=True)
        # Update flat lines threshold
        for model in models:
            model.flat_thresh = args.flat_thresh

        # Feed fits forward in temperature
        # Use T=n fit for T=n+1 guess
        for it, (model, experiment) in enumerate(zip(models, experiments)):
            # Start with fixed guess of parameters
            if it == 0:
                prev_fit = model.initial_params

            model.fit_model(
                experiment, guess=prev_fit, no_discard=args.discard_off
            )

            # Use previous fit params as next guess, if fit successful
            if len(model.fitted_params):
                prev_fit = model.fitted_params
            else:
                prev_fit = model.initial_params

        # For each field, save to file all associated fit params and model
        # funcs
        base_filename = 'ac_model_{}_1sigma_{:.5f}Oe'.format(
            models[0].NAME.replace(" ", "_"),
            dc_field
        )
        base_filename = os.path.join(user_cfg.results_dir, base_filename)
        fit_filename = '{}_params.out'.format(base_filename)
        model_filename = '{}_model.out'.format(base_filename)

        # TODO these are angular frequency (make clear in headers)
        models[0].write_fit_params(models, fit_filename)
        ac.Model.write_model_data(experiments, models, model_filename)

        if all(not model.fit_status for model in models):
            print('\n    ***Error***:')
            print('At {:.5f} Oe, no peak was measured.'.format(dc_field))
            # Exit if only one field
            if len(unique_dc_fields) == 1:
                exit()
            else:
                continue

        ac.Model.plot_colecole(
            experiments,
            models,
            save_dir=user_cfg.results_dir,
            file_head="ac"
        )
        ac.Model.plot_susceptibility(
            experiments,
            models,
            save_dir=user_cfg.results_dir,
            file_head="ac"
        )

        if args.process == 'susc':
            exit()

        all_models_sep.append(models)

    # Proceed to fit the relaxation profile after having finished fitting
    # the AC data.
    for models, experiments, dc_field in zip(all_models_sep, all_experiments_sep, unique_dc_fields): # noqa

        base_filename = 'ac_model_{}_1sigma_{:.5f}Oe'.format(
            models[0].NAME.replace(" ", "_"),
            dc_field
        )
        base_filename = os.path.join(user_cfg.results_dir, base_filename)
        fit_filename = '{}_params.out'.format(base_filename)

        print()
        print(
            '   Part II: Fit the relaxation rate temperature dependence.',
            'at {} Oe\n'.format(dc_field)
        )


        if np.sum([model.fit_status for model in models]) <= 2:
            print(
                '    ***Warning***:\n    Not enough data points to',
                'fit the relaxation profile'
            )   
            continue

        if isinstance(models[0], ac.TwoMaximaModel):
            print('\n    ***Warning***:\n')
            print(
                'Two distinct relaxation processes',
                ' detected in: {}'.format(
                    fit_filename
                )
            )
            print(
                '\n    Prepare two separate files for each process',
                ' and run RelaxationProfile --AC Generalised '
            )
            exit()

        # Instantiate the Relaxation_Profile class.
        rxp = relax.Relaxation_profile(
                verbose=args.verbose,
                sigma=args.sigma,
                exe_mode='ac',
                results_dir=user_cfg.results_dir
        )

        # Read-in the rates.
        temperature, tau = rxp.read_rates(
            fit_filename,
            args.exe_mode,
            AC_model=models[0].NAME,
            sigma=args.sigma
        )

        relax.select_model_rates(
            rxp, temperature, 1./tau, 0., _process='tau_vs_T'
        )
    return


def dc_mode_func(args):
    """
    Wrapper function for command line interface call to dc mode

    Parameters
    ----------
    args : argparser object
        command line arguments

    Returns
    -------
    None

    """

    # Check M_eq
    if args.M_eq not in ['free', 'exp', 'multiple']:
        try:
            float(args.M_eq)
        except ValueError:
            exit('\n***Error***:\n {} is invalid M_eq.\n'.format(args.M_eq))

    # Validate guess input.
    if args.guess[0] < 0.0:
        print('\n***Error***:')
        exit('Tau cannot be negative.')
    if args.model == 'single' and len(args.guess) > 1:
        print('\n***Error***:')
        exit('Single exponential model function only takes one argument (tau) as guess.') # noqa
    if args.model == 'single' and args.M_eq == 'free':
        # Dummy guess for Meq
        args.guess.append(0.1)
    if args.model == 'stretched':
        if len(args.guess) == 1 and args.M_eq != 'free':
            # Guess beta
            args.guess.append(0.95)
        if len(args.guess) == 1 and args.M_eq == 'free':
            # Guess beta and M_eq
            args.guess.append(0.95)
            args.guess.append(1.0)
        if len(args.guess) == 2 and args.M_eq == 'free':
            # Guess M_eq
            args.guess.append(1.0)
        if args.guess[1] > 1.0:
            exit('\n***Error***:\nBeta parameter cannot be larger than one.')

    # Initialise Process_DC.
    process_DC = dc.Process_DC(args)

    # Read the input file.
    data = process_DC.read_input(args.input_file, args.discard)

    # loop over the different fields.
    for field, decays in data.items():

        print('\nField: {:6.2f} Oe'.format(field))
        print('\n    Part I: Fit the magnetisation decays.')

        # Define Meq in case --M_eq Multiple is indicated.
        meq_multiple = []
        if args.M_eq == 'multiple':
            if not os.path.isfile('Multiple_Meq_{:d}Oe.txt'.format(field)):
                print('\n    ***Error***:')
                print(
                    '    --M_eq multiple specified but',
                    ' Multiple_Meq_{:d}Oe.txt is not present.'.format(
                        field
                    )
                ) # noqa
            else:
                meq_multiple = np.loadtxt(
                    'Multiple_Meq_{}Oe.txt'.format(int(field)), skiprows=1
                )

        # Define the output files.
        DC_filename, DC_params, DC_fit = ut.create_output_files(
            results_dir="{}_results".format(
                os.path.splitext(args.input_file)[0]
            ).replace(".", "_"),
            _field=field, _output='decays', exe_mode='dc',
            model_DC=args.model, M_eq=args.M_eq
            )

        # Define the model function to use.
        process_DC.define_DC_model(args)

        # Fit, plot and save the data.
        process_DC.fit_decays(
            field, decays, DC_filename, DC_params, DC_fit, args.M_eq,
            meq_multiple
        )

        # Fit the relaxation profile after fitting the DC data
        if args.process == 'all':

            print(
                '\n\n    Part II: Fit the relaxation rate',
                'temperature dependence.\n'
            )
            if len(decays[0]) <= 2:
                print(
                    '    ***Warning***:',
                    '\n    Not enough data points to ',
                    'fit the relaxation profile.'
                )
                continue

            # Instantiate the Relaxation_Profile class
            rxp = relax.Relaxation_profile(
                exe_mode='dc',
                results_dir="{}_results".format(
                    os.path.splitext(args.input_file)[0]
                ).replace(".", "_"),
                verbose=args.verbose
            )

            # Read-in the rates.
            temperature, tau = rxp.read_rates(
                DC_filename+'_params.out', exe_mode='dc',
                DC_model=args.model
            )

            # Select the model function to be used
            relax.select_model_rates(
                rxp, temperature, 1./tau, field, _process='tau_vs_T'
            )

    return


def relaxation_mode_func(args):
    """
    Wrapper function for command line interface call to relaxation mode

    Parameters
    ----------
    args : argparser object
        command line arguments

    Returns
    -------
    None

    """

    if args.process == 'tau_vs_T':
        print('\n\n    Fit the relaxation rate temperature dependence.\n')
    elif args.process == 'tau_vs_H':
        print('\n\n    Fit the relaxation rate field dependence.\n')

    rxp = relax.Relaxation_profile(
        sigma=args.sigma,
        results_dir="{}_results".format(
            os.path.splitext(args.input_file)[0]
        ).replace(".", "_"),
        verbose=args.verbose
    )

    # Read-in the rates.
    xvals, tau = rxp.read_rates(
        args.input_file, args.exe_mode, sigma=args.sigma
    )

    # Select the model function to be used
    relax.select_model_rates(
        rxp, xvals, 1./tau, args.infield, _process=args.process
    )

    return


def waveform_mode_func(args):
    """
    Wrapper function for command line interface call to waveform mode

    Parameters
    ----------
    args : argparser object
        command line arguments

    Returns
    -------
    None

    """

    # Multiple constructors implemented with call().

    # Define an empty list to store the AC susceptibilities.
    suscep = []

    # Initialise Process_Waveform
    Waveform = wfrm.Process_Waveform(args)

    # Loop over the provided data files.
    for filename in Waveform.file_names:

        # Call the initialised class with the appropriate input.
        suscep.append(
            Waveform(filename)
        )

    # Flatten the list.
    final = Waveform.flatten_and_sort(suscep)

    # Show the final susceptibilities.
    Waveform.plot_susceptibilies(final)

    # Write the input file to CCFIT
    Waveform.prepare_ccfit_AC_input(Waveform.applied_field, final)

    '''
    # Multiple constructors implemented with @singledispatchmethod decorator.
    # Initialise Process_Waveform
    Waveform = Process_Waveform(args)

    # args.dat_file is always a list. I need to figure out how to emulate
    # passing args.dat_file[0] (str) when a single file.
    Left here. It all works, just need to sort out the data type thingy
    Not added to git.

    susc = Waveform.generic_method(args.dat_file)

    # Sort the list of tuples.
    final = Waveform.sort_results(susc)

    # Show the final susceptibilities.
    Waveform.plot_susceptibilies(final)

    # Write the input file to CCFIT
    Waveform.prepare_ccfit_AC_input(Waveform.applied_field, final)
    '''

    return


def read_args(arg_list=None):
    """
    Parser for command line arguments. Uses subparsers for individual programs

    Parameters
    ----------
    args : argparser object
        command line arguments

    Returns
    -------
    None

    """

    description = """
    Program to extract the relaxation times from AC or DC data and fit their
    relaxation profile.

    Available modules:
        ccfit2 ac ...
        ccfit2 waveform ...
        ccfit2 dc ...
        ccfit2 relaxation ...
    """

    epilog = """
    To display options for a specific module, use ccfit2.py module -h
    """

    parser = argparse.ArgumentParser(
        description=description,
        epilog=epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter
        )

    # create the top-level parser
    subparsers = parser.add_subparsers(dest='exe_mode')

    # AC mode
    description_ac = """
    Extract relaxation times from AC susceptibilities using the Debye model and
    (optional) fit the resulting relaxation profile.
    """

    ac_parser = subparsers.add_parser(
        'ac',
        description=description_ac,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    ac_parser.set_defaults(func=ac_mode_func)

    ac_parser.add_argument(
        'input_file',
        metavar='<filename>',
        type=str,
        help='''File containing the AC raw data.
        See Manual for expected format.'''
    )
    ac_parser.add_argument(
        'mass',
        metavar='<value>',
        type=float,
        help='Sample mass (mg).'
    )
    ac_parser.add_argument(
        'MW',
        metavar='<value>',
        type=float,
        help='Sample molecular weight (g/mol).'
    )
    ac_parser.add_argument(
        '--process',
        metavar='<Option>',
        choices=['plot', 'susc', 'all'],
        default='all',
        help='''What to do: "plot" just shows the raw data; "susc" fits only
        the AC data; "all" fits AC data and relaxation profile.
        Options: plot, susc, all.
        Default: all.'''
    )
    ac_parser.add_argument(
        '--discard_off',
        action='store_true',
        help='Fit the susceptibilities despite not showing a peak.'
    )
    ac_parser.add_argument(
        '--temp_thresh',
        metavar='<float>',
        type=float,
        default=0.1,
        help='Threshold used to discriminate between temperatures, default 0.1 K' # noqa
    )
    ac_parser.add_argument(
        '--field_thresh',
        metavar='<float>',
        type=float,
        default=1e-8,
        help='Threshold used to discriminate between DC Fields, default 1e-8 Oe' # noqa
    )
    ac_parser.add_argument(
        '--select_T',
        action='store_true',
        help='Select the temperatures to fit.'
    )
    ac_parser.add_argument(
        '--sigma',
        metavar='<int>',
        type=int,
        default=1,
        help='log normal confidence interval. Default: 1 = 68%%.'
    )
    ac_parser.add_argument(
        '--data_pointer',
        metavar='<str>',
        type=str,
        default='[Data]',
        help=(
            'String used to locate start of data in ac.dat file '
            'Default: [Data]'
        )
    )
    # Error defined as
    # sqrt( sum(square( linear(val_x, *linear_popt) -  val_y )) ).
    # The larger this value, the tighter the constraint.\n
    ac_parser.add_argument(
        '--flat_thresh',
        metavar='<Value>',
        type=float,
        default=1E-06,
        help='Threshold to discard flat lines. Default: 1E-06.'
    )

    ac_parser.add_argument(
        '--verbose',
        action='store_true',
        help='Print the read-in values from the file indicated.'
    )

    # Hidden argument to signify that AC is being run by the executable
    # version of ccfit2
    if "--ccfit2_exe" in sys.argv:
        ac_parser.add_argument(
            '--ccfit2_exe',
            action='store_true',
            help=""
        )

    # Waveform
    description_Waveform = """
    Extract AC susceptibilities from waveform data
    See PCCP, 2019, 21, 22302-22307 for method details
    """
    epilog_Waveform = """
    This module creates a $NAME_toccfit.dat file to be passed to the AC module.
    """
    waveform_parser = subparsers.add_parser(
        'waveform',
        description=description_Waveform,
        epilog=epilog_Waveform,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    waveform_parser.set_defaults(func=waveform_mode_func)

    waveform_parser.add_argument(
        'input',
        type=str,
        metavar='<filename(s)>',
        nargs='+',
        help='SQUID file(s). Supports shell-style wildcards, e.g. data_*.dat.'
    )
    waveform_parser.add_argument(
        '--time_col',
        type=int,
        metavar='<value>',
        default=2,
        help='Column number used to read "Time stamp (s)" values. Default=2.'
    )
    waveform_parser.add_argument(
        '--temp_col',
        type=int,
        metavar='<value>',
        default=3,
        help='Column number used to read "Temperature (K)" values. Default=3.'
    )
    waveform_parser.add_argument(
        '--field_col',
        type=int,
        metavar='<value>',
        default=4,
        help='Column number used to read "Field (Oe)" values. Default=4.'
    )
    waveform_parser.add_argument(
        '--moment_col',
        type=int,
        metavar='<value>',
        default=5,
        help='Column number used to read "Moment (emu)" values. Default=5.'
    )
    waveform_parser.add_argument(
        '--field',
        type=float,
        metavar='<value>',
        default=0.,
        help='Central field value used to measure waveform data. Default=0.0.'
    )
    waveform_parser.add_argument(
        '--data_pointer',
        metavar='<str>',
        type=str,
        default='[Data]',
        help='String used to locate data in <filename(s)>. Default: "[Data]"'
    )
    waveform_parser.add_argument(
        '--field_window',
        metavar='<float>',
        type=float,
        nargs=2,
        default=[-0.5, 0.5],
        help="""Min & max field values (Oe) used to define a waveform block.
        Default: -0.5 0.5"""
    )
    waveform_parser.add_argument(
        '--show',
        action='store_true',
        help='Show individual Fourier transformed plots.'
    )
    '''
    waveform_parser.add_argument(
        '--discard_freq',
        metavar='<int>',
        type=int,
        nargs='*',
        default=[],
        help="""Outlier frequency number(s) to be discarded for each <filename>.
        You will have to inspect waveform plots and then run the program again.
        0 for None, 1 for 1st, 2 for 2nd, etc. Default: None."""
    )
    ''' # noqa

    # DC mode
    description_dc = """
    Extract relaxation times from magnetisation decays using exponentials and
    (optional) fit the resulting relaxation profile.
    """
    dc_parser = subparsers.add_parser(
        'dc',
        description=description_dc,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    dc_parser.set_defaults(func=dc_mode_func)

    dc_parser.add_argument(
        'input_file',
        metavar='<filename>',
        type=str,
        help='''File containing the DC magnetisation decay data.
        See Manual for expected format.'''
    )
    dc_parser.add_argument(
        '--process',
        metavar='<Option>',
        choices=['decays', 'all'],
        default='all',
        help=(
            'What to do: "decays" fits only the magnetisation decays; "all" '
            'continues to fit the relaxation profile. '
            'Options: {decays, all}   '
            'Default: all.'
        )
    )
    dc_parser.add_argument(
        '--model',
        metavar='<Option>',
        type=str,
        default='single',
        choices=['single', 'stretched'],
        help='''Exponential function to fit the data.
        Options: single, stretched.
        Default: single.'''
    )
    dc_parser.add_argument(
        '--M_eq',
        metavar='<Option>',
        type=str,
        default='exp',
        help='''M_eq value to use in the model.
        See Manual for options.
        Default: exp.'''
    )
    dc_parser.add_argument(
        '--guess',
        metavar='<number>',
        nargs='+',
        type=float,
        default=[400.],
        help='''Initial guess used to fit the coldest temperature.
        Number of variables depends on the indicated model.
        Default: Single (tau) -> 400; Stretched (tau, beta) -> 400, 0.95.'''
    )
    dc_parser.add_argument(
        '--discard',
        metavar='<number>',
        type=int,
        default=2,
        help='''Discard initial data points measured under a saturating DC field. # noqa
        Default: 2.'''
    )
    dc_parser.add_argument(
        '--hide_plots',
        action='store_true',
        help='''Do not show the individual mag decay plots.'''
    )
    dc_parser.add_argument(
        '--save_plots',
        action='store_true',
        help='''Save the individual mag decay plots.'''
    )
    dc_parser.add_argument(
        '--verbose',
        action='store_true',
        help='Print the read-in values from the file indicated.'
    )

    # Relaxation Profile
    description_relaxation = """
    Fit the field- and/or temperature-dependence of relaxation times.
    """
    relaxation_profile_parser = subparsers.add_parser(
        'relaxation',
        description=description_relaxation,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    relaxation_profile_parser.set_defaults(func=relaxation_mode_func)

    relaxation_profile_parser.add_argument(
        'input_file',
        type=str,
        help='File containing the relaxation times.'
    )
    relaxation_profile_parser.add_argument(
        '--process',
        metavar='<Option>',
        choices=['tau_vs_T', 'tau_vs_H'],
        default='tau_vs_T',
        help='''Fit either the temperature- or field-dependence of taus (isofield or isothermal).
        Options: tau_vs_T, tau_vs_H.
        Default: tau_vs_T.''' # noqa
    )
    relaxation_profile_parser.add_argument(
        '--infield',
        action='store_true',
        help='''Enable a Direct term in --process = tau_vs_T.'''
    )
    relaxation_profile_parser.add_argument(
        '--sigma',
        metavar='<int>',
        type=int,
        default=1,
        help='''lognorm confidence interval (AC data).
        Default: 1 = 68%%.'''
    )
    '''
    relaxation_profile_parser.add_argument(
        '--deconvolute', action='store_true',
        help='Extract two relaxation times from taus with a large associated alpha. STILL NOT AVAILABLE!'
    )
    '''# noqa
    relaxation_profile_parser.add_argument(
        '--verbose',
        action='store_true',
        help='Print the read-in values from the file indicated.'
    )
    # If argument list is none, then call function func
    # which is assigned to help function
    # read sub-parser
    parser.set_defaults(func=lambda args: parser.print_help())
    args = parser.parse_args(arg_list)
    args.func(args)

    return args


def check_version():
    try:
        response = requests.get("http://www.nfchilton.com/cc.version.cmd")
        if response.text != "3.1":
            print("\n********************************************************")
            print(
                "New version v{}".format(response.text),
                " available at www.nfchilton.com/cc-fit"
            )
            print("********************************************************\n")
    except: # noqa
        pass
    return


def main():
    read_args()



"""
This module contains functions and objects for working with DC
magnetisation decay data
"""

from math import isnan
import numpy as np
import collections
import sys
if sys.platform[0:3] == "win" or sys.platform == "darwin":
    import matplotlib
    matplotlib.use('TkAgg')
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit


# Generic dc magnetometer file header names
HEADERS_GENERIC = [
    "field",
    "temperature",
    "time",
    "moment"
]

# list of supported dc headers
# These differ between magnetometers
# One of each MUST be found in the input file
HEADERS_SUPPORTED = {
    'field': [
        'Magnetic Field (Oe)',
        'Field (Oe)'
    ],
    'temperature': [
        'Temperature (K)'
    ],
    "time": [
        'Time Stamp (sec)'
    ],
    "moment": [
        'Moment (emu)'
    ]
}


class Measurement():
    """
    Stores data for a single DC Decay measurement at a
    given temperature and applied field

    Parameters
    ----------
    dc_field : float
        Applied dc field
    temperature : float
        Temperature of experiment
    moment : float
        Magnetic moment of experiment
    time : float
        Time of experiment

    Attributes
    ----------
    dc_field : float
        Applied dc field
    temperature : float
        Temperature of experiment
    moment : float
        Magnetic moment of experiment
    time : float
        Time of experiment
    mean_temperature : float
        Mean or representative temperature assigned to this experiment
    """

    def __init__(self, dc_field, temperature, time, moment):
        self.dc_field = dc_field
        self.temperature = temperature
        self.time = time
        self.moment = moment

    def set_mean_temp(self, value: float) -> None:
        """
        Sets mean or representative value of temperature for this
        experiment

        Parameters
        ----------
        value : float
            mean value

        """
        self.mean_temperature = value

        return

    @classmethod
    def from_file(cls, file: str, header_indices: dict) -> list['Measurement']:
        """
        Extracts dc data from magnetometer output file and
        returns list of experiments, one for each valid measurment
        Incomplete lines are ignored

        Parameters
        ----------
        file : str
            Name of magnetometer output file

        header_indices : dict
            Keys are generic header names given in `HEADERS_GENERIC`
            Values are column index of header in file

        Returns
        -------
        list
            Measurement objects, one per temperature, per field
            List has the same order as the magnetometer file

        """

        # Open file and store all headers
        with open(file, 'r') as f:
            for it, line in enumerate(f):
                if '[Data]' in line:
                    header_line = it + 1

        # Columns to extract from file - n.b. these are ordered
        cols = [header_indices[gen] for gen in HEADERS_GENERIC]

        # Convert strings to floats, if not possible then mark as NaN
        converters = {
            it: lambda s: (float(s.strip() or np.NaN)) for it in cols
        }

        # Read required columns of file
        data = np.loadtxt(
            file,
            skiprows=header_line+1,
            delimiter=',',
            converters=converters,
            usecols=cols
        )

        # Convert array of floats into list of Measurement objects, one per
        # line
        # Check for missing entries which have been marked as NaN
        # and skip if found

        measurements = [
            cls(*row)
            for row in data
            if not any(isnan(val) for val in row)
        ]

        return measurements


def find_mean_temperatures(temperatures: list[float]
                           ) -> tuple[list[float], list[int]]:

    # Find values for which step size is >= 0.1 K
    mask = np.abs(np.diff(temperatures)) >= 0.1
    # and mark indices at which to split
    split_indices = np.where(mask)[0]

    # For values with similar step size, record mean temperature
    mean_temps = [
        [np.mean(grp)]*grp.size
        for grp in np.split(temperatures, split_indices + 1)
    ]

    return mean_temps, split_indices


class Experiment():
    """
    Stores data for multiple DC measurements at a
    given temperature, given applied dc field

    Parameters
    ----------
    temperature: float
        Temperature of experiment
    dc_field : float
        Applied dc field strength in Oe

    Attributes
    ----------
    temperatures: list
        Temperature of experiment
    dc_field : float
        Applied dc field strength in Oe
    """

    def __init__(self, temperatures, raw_temperatures, time, moment,
                 dc_field):

        self.temperatures = temperatures
        self.raw_temperatures = raw_temperatures
        self.time = time
        self.moment = moment
        self.dc_field = dc_field

        return

    @classmethod
    def from_measurements(cls,
                          measurements: list[Measurement]) -> list['Experiment']: # noqa

        # Sort measurements by dc field
        measurements.sort(key=lambda k: (k.dc_field, k.temperature))

        mean_temps, split_ind = find_mean_temperatures(
            [
                measurement.temperature
                for measurement in measurements
            ]
        )

        # Set each measurement's "mean", or representative temperature
        for measurement, mean_temp in zip(measurements, np.concatenate(mean_temps)): # noqa
            measurement.set_mean_temp(mean_temp)

        # sort measurements by field, then temperature
        measurements.sort(
            key=lambda k: (k.dc_field, k.mean_temperature)
        )

        raw_temperatures = [
            measurement.temperature
            for measurement in measurements
        ]

        times = np.split(
            [
                measurement.time
                for measurement in measurements
            ],
            split_ind + 1
        )

        moments = np.split(
            [
                measurement.moment
                for measurement in measurements
            ],
            split_ind + 1
        )

        dc_fields = np.split(
            [
                measurement.dc_field
                for measurement in measurements
            ],
            split_ind + 1
        )

        experiments = [
            cls(
                mean_temp, raw_temp, time, moment,
                dc_field[0]
            )
            for (
                    mean_temp, raw_temp, time,
                    moment,
                    dc_field
                ) in zip(
                    mean_temps, raw_temperatures, times,
                    moments, dc_fields
                )
        ]

        for experiment in experiments:
            print(experiment.time)

        return experiments

    @staticmethod
    def split_by_field(experiments: list['Experiment']) -> list[list['Experiment']]: # noqa

        # Split experiments into list of lists using dc field value
        # Find values for which step size is >= 1e-8 Oe
        mask = np.abs(
            np.diff([experiment.dc_field for experiment in experiments])
        ) >= 1e-8
        # and mark indices at which to split
        split_ind = np.where(mask)[0]

        experiments: list[list[Experiment]] = np.split(
            experiments, split_ind + 1
        )

        return experiments


class Read_Raw_DC_Data():
    def __init__(self, time, field, temperature, moment):
        self.time = float(time)
        self.field = float(field)
        self.temperature = float(temperature)
        self.moment = float(moment)


class Process_DC():

    def __init__(self, user_args):
        self.guess = user_args.guess
        self.hide_plots = user_args.hide_plots
        self.save_plots = user_args.save_plots

    def read_input(self, mag_decays_file, points_to_discard):
        '''
        Returns a dictionary ordered by field.
        '''

        # Data will be stored in a dictionary where the keys and values
        # will be the fields and (temperature, time, moment), respectively.
        data = {}

        with open(mag_decays_file, 'r') as f:

            # Needed to avoid UnboundLocalError: local variable 'time'
            # referenced before assignment
            temp, time, moment = None, None, None

            for line in f:

                if 'Field =' in line:

                    if time is not None:
                        # Save previous dataset
                        # TODO refactor to avoid undefined key
                        data[key][1].append(time)
                        data[key][2].append(moment)
                        # Reset
                        temp, time, moment = None, None, None

                    # initialise each value of the field (key) as a tuple of 3
                    # empty lists (temperature, time, moment)
                    key = float(line.split()[2])
                    data[key] = ([], [], [])

                elif 'T' in line:
                    data[key][0].append(float(line.split()[2]))

                    # reset lists for the current dataset
                    temp = []

                elif 'time' in line:

                    # skip the points measured with a saturating field.
                    for _ in range(points_to_discard):
                        next(f)

                    # add completed lists from previous dataset to dictionary
                    if None not in [temp, time, moment]:
                        data[key][1].append(time)
                        data[key][2].append(moment)

                    # reset lists for the current dataset
                    time = []
                    moment = []

                else:

                    time.append(float(line.split()[0]))
                    moment.append(float(line.split()[1]))

        # To collect the last data set for which there is no next(f)
        data[key][1].append(time)
        data[key][2].append(moment)

        # for key,vals in data.items():
        #    print(key, len(vals[0]), len(vals[1][0]))

        if len(data) == 0:
            print('\n***Error***:')
            print('The program could not find information on the temperature.')
            exit('Make sure the data format is: T = X K.')

        return collections.OrderedDict(sorted(data.items()))

    def define_DC_model(self, user_args):

        if user_args.model == 'single':
            self.model_name = 'single'

            if user_args.M_eq != 'free':
                self.model_function = self.single_exponential
                # tau
                self.bounds = ([0.0], [np.inf])

            else:
                self.model_function = self.single_exponential_2
                # tau, Meq
                self.bounds = ([0.0, -np.inf], [np.inf, np.inf])

        elif user_args.model == 'stretched':
            self.model_name = 'stretched'

            if user_args.M_eq != 'free':
                self.model_function = self.stretched_exponential
                # tau, beta
                self.bounds = ([0.0, 0.0], [np.inf, 1.0])

            else:
                self.model_function = self.stretched_exponential_2
                # tau, beta, Meq
                self.bounds = ([0.0, 0.0, -np.inf], [np.inf, 1.0, np.inf])
        return

    def single_exponential(self, time, tau):
        # M(t) = Meq + (M0 - Meq) * exp( -t / tau)
        return self.Meq + (self.M0 - self.Meq)*np.exp(-time/tau)

    def single_exponential_2(self, time, tau, Meq):
        # M(t) = Meq + (M0 - Meq) * exp( -t / tau)
        return Meq + (self.M0 - Meq)*np.exp(-time/tau)

    def stretched_exponential(self, time, tau, beta):
        # M(t) = Meq + (M0 - Meq) * exp( -t / tau)**beta
        return self.Meq + (self.M0 - self.Meq)*np.exp(-(time/tau)**beta)

    def stretched_exponential_2(self, time, tau, beta, Meq):
        # M(t) = Meq + (M0 - Meq) * exp( -t / tau)**beta
        return Meq + (self.M0 - Meq)*np.exp(-(time/tau)**beta)

    def fit_decays(self, field, decays, DC_filename, params_fit, vals_fit,
                   M_eq, meq_multiple):

        for it, temp in enumerate(decays[0]):

            time = decays[1][it]
            moment = decays[2][it]

            fit_failed = None

            # set initial measurement to t = 0.
            time = [i - time[0] for i in time]
            self.M0 = moment[0]
            if M_eq != 'free' and M_eq != 'exp' and M_eq != 'multiple':
                self.Meq = float(M_eq)
                if self.M0 < self.Meq:
                    print('\n***Error***:')
                    print(
                        'At {} K and {} Oe'.format(temp, field),
                        ', the provided M_eq = {} is smaller'.format(
                            self.M_eq
                        ),
                        ' than M0 = {}'.format(self.M0)
                    )
                    exit('Please revise.\n')
            elif M_eq == 'multiple':
                # check how many lines to decide how to index it.
                if np.ndim(meq_multiple) == 1:
                    self.Meq = meq_multiple[1]
                else:
                    self.Meq = meq_multiple[it, 1]
            else:
                self.Meq = moment[-1]

            if it == 0:
                guess = self.guess

            # Fit the curve and update the guess.
            # TODO: Wrap curve_fit into a class and adapt model_function(),
            # write_params_fit(), write_vals_fit() and plot()
            # to work with new class, to avoid logic with fit_failed
            try:
                popt, pcov = curve_fit(
                    self.model_function, time, moment, p0=guess,
                    bounds=self.bounds
                )
            except RuntimeError:
                fit_failed = True

            if not fit_failed:
                # compute one standard deviation errors on the parameters
                perr = np.sqrt(np.diag(pcov))
                guess = popt
                # print('    T =  {:04.2f} K; Field = {:06.2f} Oe: DONE. '.format( temp, field ), end='\r', flush=True ) # noqa

                # Calculate the model function with the optimised parameters.
                model = self.model_function(np.array(time), *popt)

                # Write the parameters of the fit.
                self.write_params_fit(params_fit, temp, field, self.M0,
                                      self.Meq, popt, perr, model, M_eq)

                # Write the values of the fit.
                self.write_vals_fit(vals_fit, temp, time, field, moment, model)

                # Plot the results.
                self.plot(temp, field, time, moment, model, popt, DC_filename)

            else:

                # Write the parameters of the fit.
                params_fit.write('{: 06.2f}  {: 06.2f} {:^10}\n'.format(
                    temp, field, 'Fit failed'
                ))

                # Write the values of the fit.
                vals_fit.write('{: 06.2f}  {: 06.2f} {:^10}\n'.format(
                    temp, field, 'Fit failed'
                ))

                pass

        params_fit.close()
        vals_fit.close()
        print('\n\n    Parameters written to')
        print('     {}_params.out'.format(DC_filename))
        print('    Fit written to')
        print('     {}_fit.out'.format(DC_filename))

        return

    def write_params_fit(self, out_obj, temperature, field, M0, Meq, popt,
                         perr, model, M_eq):

        # Calculate the ESD on tau associated with beta.
        if self.model_name == 'stretched':
            # tau, beta
            tau_ESD = self.johnston(popt[0], popt[1])

        # Write the parameters of the fit.
        if M_eq == 'free':
            out_obj.write(' {: 06.2f}  {: 06.2f}  {: 08.4E}'.format(
                temperature,
                field,
                M0
            ))
            # Meq is the last optimised parameter.
            out_obj.write(' {: 08.4E} '.format(popt[-1]))
            # exclude the last item corresponding to Meq
            for i, j in zip(popt[0:-1], perr[0:-1]):
                out_obj.write(' {: 08.4E} {: 08.4E} '.format(i, j))

        else:
            out_obj.write(' {: 06.2f}  {: 06.2f}  {: 08.4E}  {: 08.4E}'.format(
                temperature, field, M0, Meq
            ))
            for i, j in zip(popt, perr):
                out_obj.write(' {: 08.4E} {: 08.4E} '.format(i, j))

        if self.model_name == 'stretched':
            out_obj.write(' {: 08.4E} {: 08.4E}'.format(
                tau_ESD[0], tau_ESD[1]
            ))
        out_obj.write('\n')

        return

    def johnston(self, tau, beta):

        J_err_upper = tau*np.exp(
            (1.64*np.tan(np.pi*(1-beta)/2.)) / ((1 - beta)**0.141)
        )
        J_err_lower = tau*np.exp(
            -(1.64*np.tan(np.pi*(1-beta)/2.)) / ((1 - beta)**0.141)
        )

        return J_err_upper, J_err_lower

    def write_vals_fit(self, filename, temperature, time, field, moment,
                       model):

        # Write the values of the fit.
        filename.write('T = {:04.2f} K, field = {:04.2f} Oe\n'.format(
            temperature, field
        ))
        filename.write(' {:^10} {:^10} {:^10}\n'.format(
            'time (s)', 'moment', 'fit'
        ))
        for i, j, k in zip(time, moment, model):
            filename.write(' {: 08.4E} {: 08.4E} {: 08.4E}\n'.format(i, j, k))

        return

    def plot(self, temperature, field, time, moment, model, popt, DC_filename):

        if self.model_name == 'single':
            label_fit = r'$\tau={:06.2f}\ $(s)'.format(popt[0])
        elif self.model_name == 'stretched':
            label_fit = r'$\tau={:06.2f}$ (s), $\beta={:04.3f}$'.format(
                popt[0], popt[1]
            )

        # Plot the figure.
        fig, ax = plt.subplots(
            1, 1, sharex=True, sharey=True, figsize=(4.5, 3.5)
        )
        ax.plot(
            time,
            moment,
            lw=0,
            marker='o',
            fillstyle='none',
            label='{:f} K, {:f} Oe'.format(temperature, field)
        )
        ax.plot(time, model, lw=1.5, label=label_fit)

        ax.legend(loc=0, fontsize='10', numpoints=1, ncol=1, frameon=False)
        ax.set_ylabel('Moment')
        ax.set_xlabel('time (s)')

        fig.tight_layout()

        if self.save_plots:
            _filename = '{}_{}K_{}Oe_{}'.format(
                DC_filename, temperature, field, model
            )
            fig.savefig('{}.png'.format(_filename), dpi=200)
            print('\n    Plot of magnetisation decay saved as')
            print('     {}'.format(_filename))

        if not self.hide_plots:
            plt.show()

        plt.close('all')

        return
